package PIANOS.generator;
import PIANOS.Generator;
import PIANOS.datastructures.*;
import PIANOS.exceptions.*;
import PIANOS.io.*;
import java.io.*;
import java.util.*;

/**
 * This class is used to generate the Fortran module "input".
 */
public class Input {
    
    /** Generates and writes the Fortran module input.f90. 
     * @param  callParameters The correct parameter order to be used in subroutine headers
     * @param  model The model this simulation is related to
     */
    public static void generateInput(String callParameters, ComputationalModel model)
    throws IOException, SyntaxException{
        
        FortranWriter out = new FortranWriter("input.f90");
        ArrayList <String> toWrite = new ArrayList<String>();
        
        LinkedList<Variable> variables = model.getVariableList();
        LinkedList<Entity> entities = model.getEntityList();
        
        
        toWrite.add("MODULE input");
        toWrite.add("USE definitions");
        toWrite.add("IMPLICIT NONE");
        toWrite.add("");
        toWrite.add("CONTAINS");
        toWrite.addAll(generateReadData(variables, entities, callParameters));
        toWrite.addAll(generateSetInitialValues(model.getInitialValueFile(), variables, entities,
                callParameters));
        toWrite.addAll(generateReadSpatial(model.getNeighbourCount(), entities));
        toWrite.add("");
        toWrite.add("END MODULE input");
        out.write(toWrite);
    }

    // generate the subroutine read_data
    private static ArrayList<String> generateReadData(LinkedList<Variable> variables,
            LinkedList<Entity> entities, String callParameters) throws SyntaxException{
        ArrayList<String> toWrite = new ArrayList<String>();
        toWrite.add("SUBROUTINE read_data("+ callParameters +")");
        toWrite.add("! This subroutine reads the data.");
        toWrite.add("IMPLICIT NONE");
        toWrite.add("");
        // introduce the parameters.
        toWrite.addAll(Generator.generateIntroduction());
        toWrite.add("");
        toWrite.add("INTEGER :: i, j, temp, next, iostatus");
        toWrite.add("INTEGER, PARAMETER :: dp=SELECTED_REAL_KIND(8, 0)");
        toWrite.add("REAL (KIND=dp), ALLOCATABLE, DIMENSION(:) :: datafile");
        toWrite.add("INTEGER, ALLOCATABLE, DIMENSION(:) :: nexts, missingfile");
        toWrite.add("");
        toWrite.addAll(readEntityData(entities));
        toWrite.add("");
        toWrite.add("END SUBROUTINE read_data");
        toWrite.add("");
        return toWrite;
    }
    
    /* Helper methods for generateReadData
     */
    // read all entities' data
    private static ArrayList<String> readEntityData(LinkedList<Entity> entities)
    throws SyntaxException{
        ArrayList<String> toWrite = new ArrayList<String>();
        for (int i = 0; i < entities.size(); ++i) {
            if(entities.get(i).getDataFile() == null) continue;
            
            toWrite.addAll(readDataFile(entities.get(i)));
        }
        return toWrite;
    }
    
    // return correct names for datafile_1 and datafile_missing
    private static String[] getNames(String fname){
        String[] names = new String[2];
        
        if (fname.indexOf('.') < 0)
            names[0] = fname + "_1.txt";
        else
            names[0] = fname.substring(0, fname.indexOf('.')) + "_1.txt";
        
        if (fname.indexOf('.') < 0)
            names[1] = fname + "_missing.txt";
        else
            names[1] = fname.substring(0, fname.indexOf('.')) + "_missing.txt";
        
        return names;
    }
    
    private static ArrayList<String> readDataFile(Entity e) throws SyntaxException{
        ArrayList<String> toWrite = new ArrayList<String>();
        String fname = e.getDataFile();
        String data;
        String missing;
        String[] names = getNames(fname);
        data = names[0];
        missing = names[1];
        
        toWrite.add("OPEN(UNIT=16, FILE= '"+data+"', STATUS='old', IOSTAT=iostatus)");
        toWrite.addAll(ioErrorCheck("'"+data+"'", true));
        toWrite.add("OPEN(UNIT=17, FILE= '"+missing+"', STATUS='old', IOSTAT=iostatus)");
        toWrite.addAll(ioErrorCheck("'"+missing+"'", true));
        
        // file availability checked, open confirmed
        // reading left
        
        ArrayList<Variable> datavars = new ArrayList<Variable>();
        for (Variable var: e.getVariableList()){
            if (var.isData()) datavars.add(var);
        }
        
        if (datavars.size() == 1 && e.isMatrix())
            toWrite.addAll(readMatrixDataFor(data, missing, datavars.get(0)));
        else if (!e.isMatrix())
            toWrite.addAll(readDataFileFor(data, missing, datavars));
        else // matrix and multiple variables
            throw new SyntaxException("multiple data variables and matrix form entity!");
        // this HAS been checked in the parser
        
        // TODO:
        // If the array has more stuff
        // than one, generate a reading loop of the entity's size
        // and read indices to the variables according to the proto_model_spat.
        // if it has only one entry, read the whole file into the
        // variable array (easy)
        
        toWrite.add("CLOSE(17)");
        toWrite.add("CLOSE(16)");
        toWrite.add("");
        
        return toWrite;
    }
    
    private static ArrayList<String> readMatrixDataFor(String data, String missing, Variable var){
        ArrayList<String> toWrite = new ArrayList<String>();
        // ok, we have a two-dimensional variable.
        // this has been checked for validity in the parser.
        int ySize = var.getEntity().getYCoordinate().getSize();
        int xSize = var.getEntity().getXCoordinate().getSize();
        toWrite.add("next = 1");
        toWrite.add("");
        toWrite.add("DO i = 1, "+ ySize);
        toWrite.add("");
        toWrite.add("READ(UNIT=16, FMT=*, IOSTAT=iostatus) "+
                var.getName() + " % two_dim(i, :) % value");
        toWrite.addAll(ioErrorCheck("'"+data+"'"));
        toWrite.add("DO j = 1, "+ xSize);
        toWrite.add("");
        toWrite.add("IF (j == "+ xSize + ") THEN");
        toWrite.add("READ(17, FMT='(I2)', IOSTAT=iostatus) temp");
        toWrite.add("ELSE");
        toWrite.add("READ(17, FMT='(I2)', ADVANCE='NO', IOSTAT=iostatus) temp");
        toWrite.add("END IF");
        toWrite.addAll(verboseIOError("'"+missing+"'"));
        toWrite.add("IF (temp == 1) THEN");
        toWrite.add(var.getName() + " % two_dim(i, j) % is_data = .FALSE.");
        toWrite.add(var.getName() + " % two_dim_missing(next, 1) = i");
        toWrite.add(var.getName() + " % two_dim_missing(next, 2) = j");
        toWrite.add("next = next + 1");
        toWrite.add("END IF");
        toWrite.add("END DO");
        toWrite.add("END DO");
        return toWrite;
    }
    
    private static ArrayList<String> readDataFileFor(String data, String missing,
            ArrayList<Variable> datavars){
        ArrayList<String> toWrite = new ArrayList<String>();
        // the variables are all one-dimensional. let's see how long they are.
        int size = datavars.get(0).getEntity().getSize();
        toWrite.add("IF (ALLOCATED(datafile)) THEN");
        toWrite.add("DEALLOCATE(datafile)");
        toWrite.add("DEALLOCATE(missingfile)");
        toWrite.add("DEALLOCATE(nexts)");
        toWrite.add("END IF");
        toWrite.add("ALLOCATE(datafile(1:"+datavars.get(0).getEntity().getLineLength()+ "))");
        toWrite.add("ALLOCATE(missingfile(1:"+datavars.get(0).getEntity().getLineLength()+ "))");
        toWrite.add("");
        toWrite.add("ALLOCATE(nexts(1:"+datavars.size()+"))");
        toWrite.add("nexts = 1");
        toWrite.add("");
        toWrite.add("DO i = 1, "+ size);
        toWrite.add("READ(16, FMT=*, IOSTAT=iostatus) datafile(:)");
        toWrite.addAll(verboseIOError("'"+data+"'"));
        for (Variable var: datavars){
            if (var.isInteger())
                toWrite.add(var.getName() + " % one_dim(i) % value = NINT(datafile("+
                        var.getColumn() +"))");
            else
                toWrite.add(var.getName() + " % one_dim(i) % value = (datafile("+
                        var.getColumn() +"))");
        }
        toWrite.add("");
        toWrite.add("READ(17, FMT=*, IOSTAT=iostatus) missingfile(:)");
        toWrite.addAll(verboseIOError("'"+missing+"'"));
        for (int index = 0; index < datavars.size(); ++index){
            Variable var = datavars.get(index);
            toWrite.add("IF(missingfile("+ var.getColumn()+  ") == 1) THEN");
            toWrite.add(var.getName() + " % one_dim(i) % is_data = .FALSE.");
            toWrite.add(var.getName() + " % one_dim_missing(nexts("+ (index + 1) +")) = i");
            toWrite.add("nexts("+ (index +1) +") = nexts("+ (index +1) + ") +1");
            // ok, that looks horrible...
            
            toWrite.add("END IF");
            toWrite.add("");
        }
        toWrite.add("END DO");
        return toWrite;
    }
    
    
    
    // generate the subroutine set_initial_values
    private static ArrayList<String> generateSetInitialValues(String file, LinkedList<Variable>
            variables, LinkedList<Entity> entities, String callParameters) {
        ArrayList<String> toWrite = new ArrayList<String>();
        toWrite.add("");
        toWrite.add("SUBROUTINE set_initial_values("+callParameters+")");
        toWrite.add("! This subroutine reads the initial values of the parameters and sets them.");
        toWrite.add("IMPLICIT NONE");
        toWrite.add("");
        toWrite.addAll(Generator.generateIntroduction());
        toWrite.add("");
        toWrite.add("CHARACTER(LEN=100) :: line, var_name");
        // 100 should not be too shabby. This is the maximum header length. Has no
        // effect on initial value definition lines.
        toWrite.add("! HARD-CODED: the filename of the initial values file");
        toWrite.add("CHARACTER(LEN=*), PARAMETER :: initial_values_filename='"+file+"'");
        toWrite.add("INTEGER :: i, j, linenum");
        toWrite.add("INTEGER :: vertical_start, vertical_end, horizontal_start, horizontal_end");
        toWrite.add("INTEGER :: ind, iostatus");
        toWrite.addAll(readInitialValues("initial_values_filename", variables, entities));
        toWrite.add("END SUBROUTINE set_initial_values");
        toWrite.add("");
        return toWrite;
    }
    
    /* Helper methods for generateSetInitialValues
     *
     */
    
    // do the main work of initial values reading
    private static ArrayList<String> readInitialValues(String file, LinkedList<Variable> variables,
            LinkedList<Entity> entities){
        ArrayList<String> toWrite = new ArrayList<String>();
        toWrite.add("OPEN(UNIT=16, FILE=initial_values_filename, STATUS='old', IOSTAT=iostatus)");
        toWrite.addAll(ioErrorCheck(file, true));
        toWrite.add("READ (16, FMT='(A)', IOSTAT=iostatus) line");
        toWrite.addAll(verboseIOError(file));
        toWrite.add("linenum = 1");
        toWrite.add("IF (.NOT. line == '?? initial values') THEN");
        toWrite.add("WRITE(*, *) 'Expected initial values legend in ' , initial_values_filename");
        toWrite.add("WRITE(*, *) 'on line 1 not found. Quitting.'");
        toWrite.add("CLOSE(16)");
        toWrite.add("STOP");
        toWrite.add("END IF");
        toWrite.add("");
        toWrite.add("DO WHILE (.TRUE.)");
        toWrite.add("READ (16, FMT='(A)', IOSTAT=iostatus) line");
        toWrite.add("");
        toWrite.add("linenum = linenum + 1");
        toWrite.add("");
        toWrite.add("IF (iostatus < 0) THEN");
        toWrite.add("RETURN"); // finished reading
        toWrite.add("ELSE IF (iostatus > 0) THEN");
        toWrite.add("WRITE (*, *) 'Could not read ', initial_values_filename,', line ', linenum, '. Exiting program.'");
        toWrite.add("STOP");
        toWrite.add("END IF");
        toWrite.add("");
        toWrite.addAll(readVariableSpecific(variables, entities));
        toWrite.add("END DO ! Ends the file-reading loop.");
        toWrite.add("CLOSE(16)");
        return toWrite;
    }
    
    private static ArrayList<String> readVariableSpecific(LinkedList<Variable> variables,
            LinkedList<Entity> entities){
        ArrayList<String> toWrite = new ArrayList<String>();
        toWrite.add("line = TRIM(line)");
        toWrite.add("");
        toWrite.add("IF (LEN_TRIM(line) == 0) THEN");
        toWrite.add("CYCLE");
        toWrite.add("END IF");
        toWrite.add("IF (line(1:1) == '#') THEN");
        toWrite.add("CYCLE");
        toWrite.add("END IF");
        toWrite.add("");
        toWrite.add("! now we have a \"? alpha 1:something\" at hand");
        // This has been checked through and through in the parser. Doing a bit more freely here.
        toWrite.add("IF (line(1:1) /= '?')  THEN");
        toWrite.add("WRITE(*, *) 'Expected legend on line ', linenum, 'not found, quitting.'");
        toWrite.add("STOP");
        toWrite.add("END IF");
        toWrite.add("");
        toWrite.add("line = line (3:LEN_TRIM(line))");
        toWrite.add("ind = INDEX(line, ' ')");
        toWrite.add("IF (ind <= 1) THEN");
        toWrite.add("var_name = line");
        toWrite.add("ELSE");
        toWrite.add("var_name = TRIM(line(1:ind))");
        toWrite.add("END IF");
        toWrite.add("line = line(LEN_TRIM(var_name)+2:LEN_TRIM(line))");
        toWrite.add("! trim the variable name out");
        toWrite.add("! and the trailing space");
        toWrite.add("vertical_start   = 0");
        toWrite.add("vertical_end     = 0");
        toWrite.add("horizontal_start = 0");
        toWrite.add("horizontal_end   = 0");
        toWrite.add("SELECT CASE (var_name)");
        toWrite.addAll(generateInitialValueCases(variables, entities));
        toWrite.add("END SELECT");
        toWrite.add("");
        return toWrite;
    }
    
    private static ArrayList<String> generateInitialValueCases(LinkedList<Variable> variables,
            LinkedList<Entity> entities){
        ArrayList<String> toWrite = new ArrayList<String>();
        
        for (Variable v : variables){
            if (v.isFunctional())
                continue;
            if (v.isData() && v.getMissingValueCount() == 0)
                continue;
            
            toWrite.addAll(initialValueCaseFor(v));
            toWrite.add("! end of case "+v.getName());
        }
        for (Entity e: entities){
            for (Variable v: e.getVariableList()){
                if (v.isFunctional())
                    continue;
                if (v.isData() && v.getMissingValueCount() == 0)
                    continue;
                
                toWrite.addAll(initialValueCaseFor(v));
                toWrite.add("! end of case "+v.getName());
            }
        }
        toWrite.add("");
        return toWrite;
    }
    
    private static ArrayList<String> initialValueCaseFor(Variable var){
        ArrayList<String> toWrite = new ArrayList<String>();
        toWrite.add("");
        if (var.getEntity() == null)
            toWrite.addAll(initialValueCaseGlobal(var));
        else if (!var.getEntity().isMatrix())
            toWrite.addAll(initialValueCaseOneDim(var));
        else
            toWrite.addAll(initialValueCaseTwoDim(var));
        return toWrite;
    }
    
    private static ArrayList<String> initialValueCaseGlobal(Variable var){
        ArrayList<String> toWrite = new ArrayList<String>();
        String name = var.getName();
        toWrite.add("");
        toWrite.add("CASE ('"+name +"')");
        toWrite.add("READ(16, *, IOSTAT=iostatus) "+ name + " % one_dim (1) % value");
        toWrite.add("linenum = linenum + 1");
        toWrite.add("");
        toWrite.add("IF (iostatus < 0) THEN");
        toWrite.add("WRITE (*, *) 'Ran out of data when reading ', " +
                "initial_values_filename, '. Exiting program.'");
        toWrite.add("STOP");
        toWrite.add("ELSE IF (iostatus > 0) THEN");
        toWrite.add("WRITE (*, *) 'Could not read ', initial_values_filename, ', line '," +
                " linenum, '. Exiting program.'");
        toWrite.add("STOP");
        toWrite.add("END IF");
        toWrite.add("");
        toWrite.add(name +" % one_dim(1) % initial_value_set = .TRUE.");
        toWrite.add("");
        return toWrite;
    }
    
    private static ArrayList<String> initialValueCaseOneDim(Variable var){
        ArrayList<String> toWrite = new ArrayList<String>();
        String name = var.getName();
        toWrite.add("");
        toWrite.add("CASE ('"+name +"')");
        toWrite.add("! HARD-CODED:"+ name +" is one-dimensional so we consider the");
        toWrite.add("! format \""+ name + " 1:20\" for example");
        toWrite.add("ind = INDEX(line, ':')");
        toWrite.add("line(ind:ind) = ' '");
        toWrite.add("READ (line, FMT=*, IOSTAT=iostatus) vertical_start, vertical_end");
        toWrite.addAll(veryVerboseIOError("linenum", "initial_values_filename"));
        toWrite.add("DO i = vertical_start, vertical_end");
        toWrite.add("READ(16, *, IOSTAT=iostatus) "+ name +" % one_dim (i) % value");
        toWrite.add("linenum = linenum +1");
        toWrite.addAll(veryVerboseIOError("linenum", "initial_values_filename"));
        toWrite.add(name +" % one_dim(i) % initial_value_set = .TRUE.");
        toWrite.add("");
        toWrite.add("END DO");
        return toWrite;
    }
    
    private static ArrayList<String> initialValueCaseTwoDim(Variable var){
        ArrayList<String> toWrite = new ArrayList<String>();
        String name = var.getName();
        toWrite.add("");
        toWrite.add("CASE ('"+name +"')");
        toWrite.add("! HARD-CODED: "+ name +" is two-dimensional so we consider the");
        toWrite.add("! format \""+ name +" 1:20 5:400\" for example");
        toWrite.add("DO i = 1, 2");
        toWrite.add("ind = INDEX(line, ':')");
        toWrite.add("line(ind:ind) = ' '");
        toWrite.add("END DO");
        toWrite.add("");
        toWrite.add("READ (line, FMT=*, IOSTAT=iostatus) vertical_start, vertical_end, horizontal_start, horizontal_end");
        toWrite.add("");
        toWrite.addAll(veryVerboseIOError("linenum", "initial_values_filename"));
        toWrite.add("DO i = vertical_start, vertical_end");
        toWrite.add("READ(16, *, IOSTAT=iostatus) "+ name +" % two_dim (i, horizontal_start:horizontal_end) % value");
        toWrite.add("linenum = linenum +1");
        toWrite.addAll(veryVerboseIOError("linenum", "initial_values_filename"));
        toWrite.add(name +" % two_dim(i, horizontal_start:horizontal_end) % initial_value_set = .TRUE.");
        toWrite.add("");
        toWrite.add("END DO");
        //TODO: UNDER CONSTRUCTION
        return toWrite;
    }
    
    // generate the subroutine read_spatial
    private static ArrayList<String> generateReadSpatial(int neighbours,
            LinkedList<Entity> entities) {
        ArrayList<String> toWrite = new ArrayList<String>();
        Entity spatial = null;
        for (int i = 0; i < entities.size(); ++i){
            if (entities.get(i).isSpatial()){
                spatial = entities.get(i);
                break;
            }
        }
        if (spatial == null){
            // There are no spatial entities in this model
            return toWrite;
        }
        int size = spatial.getSize();
        toWrite.add("");
        toWrite.add("SUBROUTINE set_spatial(spatial)");
        toWrite.add("! sets the array describing neighbour relations");
        toWrite.add("! according to the adjacency matrix file.");
        toWrite.add("IMPLICIT NONE");
        toWrite.add("");
        toWrite.add("! HARD-CODED: the dimensions (see main)");
        
        toWrite.add("INTEGER, DIMENSION("+ size +", "+ (neighbours +1) +"), INTENT(OUT) :: spatial");
        toWrite.add("INTEGER, DIMENSION("+size +", "+ size +") :: matrix");
        toWrite.add("INTEGER :: iostatus, i, j, num");
        toWrite.add("CHARACTER (LEN=*), PARAMETER :: filename='"+ spatial.getSpatialMatrixFile() +"'");
        toWrite.add("");
        toWrite.add("! HARD-CODED: the name of the adjacency matrix file ");
        toWrite.add("");
        toWrite.add("! Reads in the adjacency matrix.");
        toWrite.add("");
        toWrite.add("OPEN(UNIT=16, FILE=filename, STATUS='old', IOSTAT=iostatus)");
        toWrite.addAll(ioErrorCheck("filename", true));
        toWrite.add("DO i=1, "+ size);
        toWrite.add("READ(16, FMT=*, IOSTAT=iostatus) matrix(i, :)");
        toWrite.addAll(ioErrorCheck("filename"));
        toWrite.add("END DO");
        toWrite.add("");
        toWrite.add("DO i = 1, "+ size);
        toWrite.add("num = 0");
        toWrite.add("DO j = 1, "+ size);
        toWrite.add("IF (matrix(i, j) == 1) THEN");
        toWrite.add("IF (num > "+ neighbours +") THEN");
        toWrite.add("WRITE (*, *) 'More neighbours than expected for (', i, j, '), exiting.'");
        toWrite.add("STOP");
        toWrite.add("END IF");
        toWrite.add("spatial(i, num + 2) = j");
        toWrite.add("num = num + 1");
        toWrite.add("END IF");
        toWrite.add("END DO");
        toWrite.add("spatial(i, 1) = num");
        toWrite.add("END DO");
        toWrite.add("");
        toWrite.add("END SUBROUTINE set_spatial");
        toWrite.add("");
        return toWrite;
    }
    
    /* General help methods
     */
    
    // checks for errors and prints a very verbose io error message
    private static ArrayList<String> veryVerboseIOError(String linevar, String fname){
        ArrayList<String> toWrite = new ArrayList<String>();
        toWrite.add("");
        toWrite.add("IF (iostatus < 0) THEN");
        toWrite.add("WRITE (*, *) 'Ran out of data when reading line ', "+ linevar +", ' in ', "+ fname +", '. Exiting program.'");
        toWrite.add("STOP");
        toWrite.add("ELSE IF (iostatus > 0) THEN");
        toWrite.add("WRITE (*, *) 'Could not read ', "+ fname +", ', line ', "+ linevar +", '. Exiting program.'");
        toWrite.add("STOP");
        toWrite.add("END IF");
        toWrite.add("");
        return toWrite;
    }
    
    // checks for errors and prints a verbose io error message
    private static ArrayList<String> verboseIOError(String fname){
        ArrayList<String> toWrite = new ArrayList<String>();
        toWrite.add("");
        toWrite.add("IF (iostatus < 0) THEN");
        toWrite.add("WRITE (*, *) 'Ran out of data when reading ', "+ fname +", '. Exiting program.'");
        toWrite.add("STOP");
        toWrite.add("ELSE IF (iostatus > 0) THEN");
        toWrite.add("WRITE (*, *) 'Could not read ', "+ fname +", '. Exiting program.'");
        toWrite.add("STOP");
        toWrite.add("END IF");
        toWrite.add("");
        return toWrite;
    }
    
    /** runs ioErrorCheck with fname and false.
     */
    private static ArrayList<String> ioErrorCheck(String fname){
        return ioErrorCheck(fname, false);
    }
    /** generate: check if iostatus is wrong, and print an error about fname
     * if open == false, say that could not read, if open == true, say could not open
     **/
    private static ArrayList<String> ioErrorCheck(String fname, boolean open){
        ArrayList<String> toWrite = new ArrayList<String>();
        toWrite.add("");
        toWrite.add("IF (iostatus /= 0) THEN");
        if (open)
            toWrite.add("WRITE (*, *) 'Could not open ', "+fname+", '. Exiting program.'");
        else
            toWrite.add("WRITE (*, *) 'Could not read ', "+fname+", '. Exiting program.'");
        toWrite.add("STOP");
        toWrite.add("END IF");
        toWrite.add("");
        return toWrite;
    }
}
