package PIANOS.io;

import java.io.*;

import java.util.ArrayList;
import java.util.regex.*;

/**
 * FortranWriter, a class that receives lines of Fortran code, multilines and indents them correctly and writes them to a file.
 */
public class FortranWriter {
    
    private static final String[] BEGIN_INDENT = {"BLOCK\\p{Space}*DATA", "DO", "FORALL",
						  "FUNCTION", "IF", "INTERFACE", "MODULE", "TYPE\\p{Space}*[.[^\\(]]+", "PROGRAM",
						  "SELECT", "SUBROUTINE", "CASE\\p{Space}*\\(", "WHERE", "CONTAINS", "ELSE", "ELSEWHERE"};
            
    private static final String[] END_INDENT = {"END", "CASE\\p{Space}*\\(", "CONTAINS", "ELSE", "ELSEWHERE"};
            
    private static final int INDENT_CHARS = 4;
    private static final int MAX_LINE_LENGTH = 79;
            
    private File file;
    private int indent;
    private boolean first_write = true;
            
    /**
     * Creates a new <code>FortranWriter</code>.
     *
     * @param   fileName    name of the file the Fortran source code will be
     *                      written to
     */
    public FortranWriter(String fileName) {
	file = new File(fileName);
	indent = 0;
    }
            
    /* 
       //Considered obsolete and slow
       //Call write(ArrayList<String>) or write(String[]) 
       //so that the file is not closed and opened every time

       public void write(String line) throws IOException {
       String[] array = new String[1];
       array[1] = line;
       this.write(array);
       }
    */      
    /**
     * Writes lines of Fortran source code into the file.
     *
     * @param   lines   an ArrayList containing the lines to be written to the
     *                  file
     *
     * @throws  IOException If the file couldn't be written to
     */
            
    public void write(ArrayList<String> lines) throws IOException{
	write(lines.toArray(new String[lines.size()]));
    }
            
    /**
     * Writes lines of Fortran source code into the file.
     *
     * @param   lines   an array containing the lines to be written to the
     *                  file
     *
     * @throws  IOException If the file couldn't be written to
     */
    public void write(String[] lines) throws IOException {
	FileWriter out;
	if (first_write) {
	    out = new FileWriter(file, false);
	    first_write = false;
	} else out = new FileWriter(file, true);
                
	for (int j = 0; j < lines.length; j++){
	    String line = lines[j].trim();
                    
	    if (line.length() <= 0){
		out.write(getIndent() + line + "\n");
		continue;
	    }
	    if (line.charAt(0) == '!'){
		out.write(getIndent() + line + "\n");
		continue;
	    }
                    
	    for (int i = 0; i < END_INDENT.length; i++)
		if (line.matches("^\\p{Space}*" + END_INDENT[i] + ".*$"))
		    indent = indent - INDENT_CHARS;
                    
	    String  loopline = line; // we must save the original line.
                    
	    if (indent + loopline.length() > MAX_LINE_LENGTH) {
		boolean firstLine = true;
		boolean characterContinue = false;
		String temp = loopline;
		while (temp != null){
		    if (indent + temp.length() > MAX_LINE_LENGTH) {
			if (!firstLine)
			    indent = indent + INDENT_CHARS;
                                
			int indentedMax = MAX_LINE_LENGTH -indent;
                                
			//int last_space = temp.lastIndexOf(" ", MAX_LINE_LENGTH - 2);
			int last_space = temp.lastIndexOf(" ", indentedMax - 2);
                        if (last_space == -1){
                            last_space = temp.lastIndexOf(" ");
                            if (last_space == -1){
                            System.out.println("In "+ file + ": Please cut the line \"" +line + "\" manually " +
                                    "if You have problems compiling the program. " +
                                    "I can't see how to cut it properly");
                        break; // if we can't cut it let's leave it and hope
                            }
                        } 
                        
			out.write(getIndent() + temp.substring(0, last_space) + " &\n");
			// PROBLEM. MESSES UP CHARACTER STRINGS.
			// FIX. regex for character strings in Fortran:
			// '[.[^']]*'
			// This is assuming that '' is not used, and " are not used.
			// check if last_space is inside a matching section.
			Pattern p = Pattern.compile("'[.[^']]*'");
			Pattern cl = Pattern.compile("^\\&[.[^']]*'");
			Matcher ch = cl.matcher(temp);
			Matcher m = p.matcher(temp);
			boolean matched = false;
			while (m.find()){
			    String characters = m.group();
			    if  (m.start() <= last_space && last_space < m.end()){
				// it's inside. We need a & after last_space.
				temp = temp.substring(0, last_space + 1) + "&" + temp.substring(last_space +1);
				matched = true;
			    }
			}
			if (!matched){
			    while (ch.find()){
				String characters = ch.group();
				if  (ch.start() <= last_space && last_space < ch.end()){
				    // it's inside. We need a & after last_space.
				    temp = temp.substring(0, last_space + 1) + "&" + temp.substring(last_space +1);
				}
                                        
			    }
			}
			// we did not skew index numbering, hopefully.
			loopline = temp.substring(last_space + 1);
			temp = loopline;
                                
			if (!firstLine)
			    indent = indent - INDENT_CHARS;
			else firstLine = false;
		    } else {
			indent = indent + INDENT_CHARS;
			out.write(getIndent() + temp + "\n");
			indent = indent - INDENT_CHARS;
			// also indent the last continual line
			temp = null;
			// only after THAT terminate the loop.
                        /* in Anni's version the last continual line was
                         * mistakenly lost.
                         */
		    }
                            
		}
	    } else out.write(getIndent() + line + "\n");
                    
                    
	    for (int i = 0; i < BEGIN_INDENT.length; i++){
		if (line.matches("^\\p{Space}*" + BEGIN_INDENT[i] + ".*$")){
		    if (!line.matches("^\\p{Space}*END.*$") &&
			!line.matches("^\\p{Space}*MODULE\\p{Space}+PROCEDURE\\p{Space}+.*$") &&
			!line.matches("^\\p{Space}*MODULE\\p{Space}+procedure\\p{Space}+.*$")) {
			indent = indent + INDENT_CHARS;                               
		    }
		}
	    }
	}
	out.close();
    }
            
    private String getIndent() {
	String spaces = "";
	for (int i = 0; i < indent; i++) {
	    spaces = spaces + " ";
	}
	return spaces;
    }
            
}
