! Prototype program for sequential
! update of variables, without blocking
! with a single total iterations option
! of a simple model, where

! alpha is a global parameter gauging the influence of northerness on 
! the bird species 
! northerness is a parameter that represents the northerness of a square
! (middle line drawn in the middle of Finland)
! p is the probability of a bird living in a square; 
! also x and obs are variables of "bird", "square"-pairs.
!
! alpha conforms to the uniform continuous distribution from -1 to 1
! p is a functional variable, namely:
!
! p_i_j = exp(alpha * northerness_i) / (1 + exp(alpha * northerness_i))
!
! x conforms to Bernoulli(p)
!
! obs conforms to a user-defined distribution:
!
! P(obs = 0 | x = 0) = 0.8
! P(obs = 1 | x = 0) = 0.2
! P(obs = 0 | x = 1) = 0.4
! P(obs = 1 | x = 1) = 0.6
!
!  ASCII graphics:
!
! alpha ..... >   p  ----->  x ----> obs
!
!                ^
!              .
! northerness .  
!
!
!
! reads: initial values, data, adjacency matrix
!
! 
!
!


PROGRAM main
  USE definitions
  USE input
  USE output
  USE proposal
  ! THIS is a hardcoded default location for user distributions.
  USE user_dist
  IMPLICIT NONE 
  
  ! HARD-CODED: the variables in the model and their type (real or integer)
  TYPE(variables_real) :: alpha ! global variable
  TYPE(variables_int) :: northerness
  
  TYPE(variables_int) :: x
  TYPE(variables_int) :: obs
  TYPE(variables_int) :: q
  TYPE(variables_real) :: p
  
 ! the spatial structure
  
  INTEGER, DIMENSION(21, 3) :: spatial
  ! HARD-CODED:
  ! first dimension: the number of spatial units (squares)
  ! second dimension: the number of maximum neighbours + 1
  
  ! the spatial(i, 1) defines how many neighbours square i has and 
  ! spatial(i, 2) ... spatial(i, n) define their indices.

  EXTERNAL G05CBF, G05ZAF


  ! HARD-CODED: the output file name
  CHARACTER(LEN=*), PARAMETER :: output_filename='output_values.txt'

  INTEGER :: i, progress
  INTEGER, PARAMETER :: multiplier=100 
  ! this is the length multiplier 
  ! for a proposal buffer, for example if alpha has 21 values, then 
  ! the buffer for alpha would be 100 * 21 = 30 000 in size.

  ! HARD-CODED: The thinning factor and the burn-in iteration count
  INTEGER, PARAMETER :: thinning=100, burn_in=100

  
  CALL random_init()
  

  ! these are left to be defined later
  ALLOCATE(alpha % one_dim(1: 1))
  ALLOCATE(alpha % buffer(multiplier))

  ! HARD-CODED: the dimension (21 squares)
  ALLOCATE(northerness % one_dim(1: 21))
  
  ! HARD-CODED: the dimension (birds-horizontal, squares-vertical)
  ALLOCATE(q % two_dim(1:21, 1:4))
  ALLOCATE(p % two_dim(1:21, 1:4))

  ALLOCATE(x % two_dim(1:21, 1:4))
  ALLOCATE(x % buffer(21*4 * multiplier))

  ALLOCATE(obs % two_dim(1:21 , 1:4))
  
  ! then set all updates to zero.
  ! wanted updates are not set, this time the user has requested a total
  ! update number to be done.

  alpha % one_dim  % successful_changes = 0
  alpha % buffer_index = -1
  
  northerness  % one_dim % successful_changes = 0
  northerness  % buffer_index = -1
  
  x     % two_dim % successful_changes = 0
  x     % buffer_index = -1
  
  ! setting the initial values and reading data.
  CALL read_data(alpha, northerness, q, p, x, obs)
  CALL set_initial_values(alpha, northerness, q, p, x, obs)
  CALL set_functional(alpha, northerness, q, p, x, obs)
  CALL set_spatial(spatial)
  
  OPEN(UNIT=15, FILE=output_filename)

  ! measures the program execution progress. 
  ! each step is one percent of execution, that is
  ! progress = iterations / 100
  ! or in the case of random update strategy, it is
  ! progress = NINT(REAL(variables_achieved)/total * 100.0)
  progress = 0


! HARD-CODED: the sequential update strategy 
! HARD-CODED: the number of iterations (here 1000)
  DO i=1, 10000
     ! if a soft-stop key is detected, EXIT
     
     CALL update_all(alpha, northerness, q, p, x, obs)
     
     IF (i > burn_in .AND. MOD(i, thinning) == 0) THEN
        CALL write_output(alpha, northerness, q, p, x, obs)
     END IF

     IF(i/100 > progress) THEN
        progress = i/100
        WRITE(*, *) progress, '% of simulation complete'
     END IF
  END DO
  
  
  CALL write_summary(alpha, northerness, q, p, x, obs)
  CALL write_last_values(alpha, northerness, q, p, x, obs)

  CLOSE(15)
  WRITE(*, *) 'Simulation program completed successfully'


CONTAINS    
  SUBROUTINE random_init()
    ! This subroutine initializes the default random number generator.
    IMPLICIT NONE
!    EXTERNAL G05CBF, G05ZAF
    INTEGER :: size, status
    INTEGER, DIMENSION(8) :: time
    INTEGER, DIMENSION(:), ALLOCATABLE :: seed
    
    CALL RANDOM_SEED(size=size)
    ALLOCATE(seed(size), stat=status)
    IF (status > 0) STOP 'Failed to allocate!'
    CALL DATE_AND_TIME(values = time)
    seed = 100*time(7) + time(8)/10
    CALL G05ZAF('O')
    CALL G05CBF(seed)
    
  END SUBROUTINE random_init
  
  SUBROUTINE set_functional(alpha, northerness, q, p, x, obs) 

    TYPE(variables_real), INTENT(INOUT) :: alpha
    TYPE(variables_int), INTENT(INOUT) :: northerness
    
    TYPE(variables_int), INTENT(INOUT) :: q
    TYPE(variables_real), INTENT(INOUT) :: p   
    TYPE(variables_int), INTENT(INOUT) :: x
    TYPE(variables_int), INTENT(INOUT) :: obs  
    INTEGER :: sum, i, j, k

    DO i = 1, 21
       DO j = 1, 4
          sum = 0
          ! + 1 because the first index is the number of neighbours,
          ! the rest are the neighbouring indices of squares
          ! updating q % two_dim(spatial(i, k), j)
          DO k = 1, spatial(i, 1)
             sum = sum + x % two_dim(spatial(i, k + 1), j) % value
          END DO
          q % two_dim(i, j) % value = sum
       END DO
    END DO

  END SUBROUTINE set_functional
  
  SUBROUTINE update_all(alpha, northerness, q, p, x, obs)    
    ! This subroutine updates all the parameters once.
    
    ! assumed: 
    ! initial values have been set, data read into memory 
    
    IMPLICIT NONE
    
    
    TYPE(variables_real), INTENT(INOUT) :: alpha
    TYPE(variables_int), INTENT(INOUT) :: northerness
    
    TYPE(variables_int), INTENT(INOUT) :: q
    TYPE(variables_real), INTENT(INOUT) :: p   
    TYPE(variables_int), INTENT(INOUT) :: x
    TYPE(variables_int), INTENT(INOUT) :: obs    
    
    INTEGER, PARAMETER :: dp = SELECTED_REAL_KIND(8, 0)

    INTEGER :: i, j, k, prob_loop, sum, h, neighbour
    REAL (KIND=dp) :: p_acc, new_frequency, frequency
    
    INTEGER :: int_transition
    REAL (KIND=dp) :: real_transition
        
    ! NAG wants the double precision.
    REAL (KIND=dp), SAVE, DIMENSION(1000) :: prob
    INTEGER, SAVE :: prob_index = -1
    
    EXTERNAL G05FAF

!    WRITE(*, *) 'Debug: about to generate random numbers'

    IF (prob_index > SIZE(prob) .OR. prob_index < 1) THEN
!       DO prob_loop = 1, SIZE(prob)
!          WRITE(*, *) 'Debug: just about to generate'
          CALL G05FAF(0.0_dp, 1.0_dp, SIZE(prob), prob)
!       END DO
       prob_index = 1       
    END IF
    
!    WRITE(*,*) 'Debug: randoms generated'
    !    updating the single alpha
    
    p_acc=1
    IF (alpha % buffer_index > SIZE(alpha % buffer) &
         .OR. alpha % buffer_index < 1) THEN
       CALL generate('alpha', alpha % buffer)
       alpha % buffer_index = 1
    END IF
    
    alpha % one_dim(1) % new_value = alpha % buffer(alpha % buffer_index)
    alpha % buffer_index = alpha % buffer_index +1
    
    ! kuolema. pit katsoa alphan priorijakaumaa ja sen lisksi p:n
    ! kautta x:n jakaumaa.
    
    CALL continuous_uniform_freq(-1.0_dp, 1.0_dp, alpha % one_dim(1) % value, frequency)
    CALL continuous_uniform_freq(-1.0_dp, 1.0_dp, alpha % one_dim(1) % new_value, new_frequency)
    
    p_acc=p_acc*new_frequency/frequency
    
    ! HARD-CODED: alpha affects p which is a functional parameter
    ! So we need to get a new value for p and calculate the likelihood
    ! of x given the new value of p.
    
    DO i=1, 21 ! the number of squares
       DO j = 1, 4 ! birds
          p % two_dim(i,j) % value = &
               EXP(alpha % one_dim(1) % value * northerness % one_dim(i) % value + q % two_dim(i, j) % value)/(1 + EXP(alpha % one_dim(1) % value * northerness % one_dim(i) % value + q % two_dim(i, j) % value)) 
          ! (do we need this.. or do we assume that the current
          ! value is always correct)
          ! well, this way it's surely correct.
          
          p % two_dim(i,j) % new_value = &
               EXP(alpha % one_dim(1) % new_value * northerness % one_dim(i) % value + q % two_dim(i, j) % value)/ & 
               (1 + EXP(alpha % one_dim(1) % new_value * northerness % one_dim(i) % value + q % two_dim(i, j) % value))
       END DO
    END DO
    
    DO i=1, 21 ! number of squares
       DO j=1, 4 ! number of birds
          
          ! ok, how probable is THE MODEL if we change alpha a little...
          ! calculating the likelihood of ALL the x in this loop.
          
          CALL bernoulli_freq(p % two_dim(i, j) % value, &
               x % two_dim(i, j) % value, frequency)
          CALL bernoulli_freq(p % two_dim(i, j) % new_value, &
               x % two_dim(i, j) % value, new_frequency) 
          
          p_acc=p_acc*new_frequency/frequency
          
       END DO
    END DO
    
    ! P(alfa') * P(X | alpha', northerness)  / the old value
    ! This has now been calculated
    
    ! q(a', a) / q(a, a') remains
    
    ! HARD-CODED: the proposal distribution of alpha
    ! alpha's proposal distribution was .
    
    CALL continuous_uniform_freq(-1.0_dp, 1.0_dp, alpha % one_dim(1) % value, frequency)
    CALL continuous_uniform_freq(-1.0_dp, 1.0_dp, alpha % one_dim(1) % new_value, new_frequency)
    
    p_acc = p_acc * frequency/new_frequency
    
    ! this is the M-H acceptation...
    IF (p_acc > prob(prob_index)) THEN
       alpha % one_dim(1) % value = alpha % one_dim(1) % new_value 
       alpha % one_dim(1) % successful_changes = alpha % one_dim(1) % successful_changes + 1
       
       DO i=1, 21 ! the number of squares
          DO j = 1, 4 ! birds
             p % two_dim(i,j) % value = p % two_dim(i,j) % new_value
          END DO
       END DO
    END IF
    
    
    prob_index = prob_index + 1
    
    
    IF (prob_index > SIZE(prob) .OR. prob_index < 1) THEN
!          WRITE(*, *) 'Debug: about to generate '
          CALL G05FAF(0.0_dp, 1.0_dp, SIZE(prob), prob)
       prob_index = 1       
    END IF

    
    ! END OF UPDATE alpha.
  
    ! from this point unaltered copied information, needs to be made to comply
    ! to the new model
    
    ! updating x
    DO i = 1, 21
       DO j = 1, 4
          p_acc = 1
          
          IF (x % buffer_index > SIZE(x % buffer) .OR. x % buffer_index < 1) THEN
             CALL generate('x', x % buffer)
             x % buffer_index = 1
          END IF
          
          ! this is new value generation on a fixed proposal strategy, 
          ! the generator has chosen this according to instructions
          
          ! HARD CODED: x's proposal strategy (fixed or rw, fixed chosen)
          x % two_dim(i, j) % new_value = x % buffer(x % buffer_index)
          
          x % buffer_index = x % buffer_index + 1
          
          ! now x has a new value.
          
          ! here we have hardcoded the formula for x's likelihood.
          
          ! HARD-CODED: how to calculate x's accepting probability
          
          ! this is: p(x')
          
     !     WRITE(*, *) 'Debug: calling bernoulli with ', & 
     !          p % two_dim(i, j) % value, x % two_dim(i, j) % new_value, new_frequency
          CALL bernoulli_freq(p % two_dim(i, j) % value, & 
               x % two_dim(i, j) % new_value, new_frequency)
          ! this is: p(x)
          CALL bernoulli_freq(p % two_dim(i, j) % value, & 
               x % two_dim(i, j) % value, frequency)
        
          p_acc = p_acc * new_frequency/frequency
          
          ! this density function is assumed to first take the parameters 
          ! and then the point for which the density is computed.
          
          
          ! HARD-CODED: obs depends on x:
          
          ! this is p(obs | x')
          CALL defined_points_freq(x % two_dim(i, j) % new_value, &
               obs % two_dim(i, j) % value, new_frequency)
          
          CALL defined_points_freq(x % two_dim(i, j) % value, &
               obs % two_dim(i, j) % value, frequency)
          
          p_acc = p_acc * new_frequency / frequency
          
          ! The q(alpha', alpha) / q(alpha, alpha')
          ! if fixed, this is P(alpha) / P(alpha')
          
          ! HARD-CODED: the proposal distribution of x
          ! 
          ! this is easy:
          
          
          CALL discrete_uniform_freq(0, 1, x % two_dim(i, j) % new_value, new_frequency)
          CALL discrete_uniform_freq(0, 1, x % two_dim(i, j) % value, frequency)
          
          p_acc = p_acc * frequency/new_frequency
          
          ! this is the M-H acceptation...
          IF (p_acc > prob(prob_index)) THEN
             x % two_dim(i, j) % value = x % two_dim(i, j) % new_value 
             x % two_dim(i, j) % successful_changes = x % two_dim(i, j) % successful_changes + 1
          END IF
          
          
          
          prob_index = prob_index + 1
          
          IF (prob_index > SIZE(prob) .OR. prob_index < 1) THEN
!             WRITE(*, *) 'Debug: about to generate'
             CALL G05FAF(0.0_dp, 1.0_dp, SIZE(prob), prob)
             prob_index = 1       
          END IF
          
          ! 1 to max number of neighbours
          DO k = 1, spatial(i, 1)
             sum = 0
             ! + 1 because the first index is the number of neighbours,
             ! the rest are the neighbouring indices of squares
             neighbour = spatial(i, k + 1)
             ! updating q % two_dim(spatial(i, k), j)
             DO h = 1, spatial(neighbour, 1) 
                sum = sum + x % two_dim(spatial(neighbour, h + 1), j) % value
             END DO
             q % two_dim(neighbour, j) % value = sum
          END DO

       END DO ! End of update all X.
    END DO
    
  END SUBROUTINE update_all
  
END PROGRAM main
