package PIANOS;

import PIANOS.datastructures.*;
import PIANOS.exceptions.*;
import PIANOS.io.*;
import PIANOS.io.ComputationalModelParser;
import PIANOS.generator.*;

import java.util.*;
import java.io.*;

/** The PIANOS main class. 
 Writes Fortran module files for a simulation program that matches the model description files given. 
 May be run on the command line with 8 or 1 parameters, see the PIANOS manual on the
 CD-ROM or in the tar.gz for details.
 See writeProgram for the Java way of running PIANOS.
 */

public class Generator{
    private static ComputationalModel model = null;
    private static DistributionFactory factory = null;
    private static FortranWriter writer;
    
    //For testing...
    public static void setModel(ComputationalModel modelToSet){
        model = modelToSet;
    }
    
    /** The PIANOS main interface. Writes definitions.f90, input.f90, output.f90, proposal.f90, main.f90. 
            Note that when compiling the user will also need user_dist.f90 that contains the user-defined
            distributions used in the model.
     
        @param userDistFileName the filename of the user-defined distributions file
        @param modelFileName the filename of the model description file
        @param initialValueFileName the filename of the simulation initial values file
        @param simulationFileName the filename of the simulation parameters file (burn-in, thinning)
        @param proposalFileName the filename of the proposal strategies and distributions file
        @param updateFileName the filename of the update strategy and iterations file
        @param toOutputFileName the filename of the parameters desired as output
        @param lastValuesFileName the filename of the file where the last values are to be saved after the simulation
     */
    public static void writeProgram(String userDistFileName, String modelFileName,
            String initialValueFileName, String simulationFileName, String proposalFileName,
            String updateFileName, String toOutputFileName, String lastValueFileName)
            throws MissingFunctionException, SyntaxException, InvalidProposalException, IOException,
            MissingDistributionException, IllegalParametersException,
            InvalidModelException{
        
	

        model = null;
        factory = new DistributionFactory(userDistFileName);
        
        
        model = ComputationalModelParser.readModel(modelFileName, initialValueFileName,
                simulationFileName, proposalFileName, updateFileName, toOutputFileName, factory);
        
	System.out.print("Writing simulation program...");
        
        
        model.setLastValuesFileName(lastValueFileName);
        
        LinkedList<Variable> variables = model.getVariableList();
        
        // DEBUG
        /*
        System.out.println("GENERATOR DEBUG: Generator.java: Global variable list from the parser");
        for (Variable v: variables)
            System.out.println(v.getName());
        */

        String callParameters = setCallParameters(model);
        
        Definitions.generateDefinitions(model);
        
        // DEBUG
        /*
        variables = model.getVariableList();
        System.out.println("GENERATOR DEBUG: Generator.java: Global variable list after Definitions");
        for (Variable v: variables)
            System.out.println(v.getName());
        */

        Input.generateInput(callParameters, model);
        
        // DEBUG
        /*
        variables = model.getVariableList();
        System.out.println("Generator.java: Global variable list after Input");
        for (Variable v: variables)
            System.out.println(v.getName());
        */

        Output.generateOutput(callParameters, model);
        
        // DEBUG
        /*
        variables = model.getVariableList();
        System.out.println("Generator.java: Global variable list after Output");
        for (Variable v: variables)
            System.out.println(v.getName());
        */

        Proposal.generateProposal(model);
        
        // DEBUG
        /*
        variables = model.getVariableList();
        System.out.println("Generator.java: Global variable list after Proposal");
        for (Variable v: variables)
            System.out.println(v.getName());
        */

        FortranMain.generateMain(callParameters, model);
        
        // DEBUG
        /*
        variables = model.getVariableList();
        System.out.println("GENERATOR DEBUG: Generator.java: Global variable list after FortranMain");
        for (Variable v: variables)
            System.out.println(v.getName());
        */
	System.out.println("\t\t\t\t\t[OK]");
        System.out.println("The program completed successfully.");
    }
    
    /** The PIANOS command-line interface. Writes definitions.f90, input.f90, output.f90, proposal.f90, main.f90. 
            Note that when compiling the user will also need user_dist.f90 that contains the user-defined
            distributions used in the model. Command-line parameters specified in order 
            (choose first 8 or the last one which incorporates the 8):
     
        @param userDistFileName the filename of the user-defined distributions file
        @param modelFileName the filename of the model description file
        @param initialValueFileName the filename of the simulation initial values file
        @param simulationFileName the filename of the simulation parameters file (burn-in, thinning)
        @param proposalFileName the filename of the proposal strategies and distributions file
        @param updateFileName the filename of the update strategy and iterations file
        @param toOutputFileName the filename of the parameters desired as output
        @param lastValuesFileName the filename of the file where the last values are to be saved after the simulation
        @param allFilesFileName (mutually exclusive with the parameters above)
    A file containing the file names of all the files specified above, in the format: <br>
    <i>(begin of file)</i><br>
    user distributions:     relative/path/to/user_dist.f90 <br>
    model:                  relative/path/to/modelFile<br>
    initial values:         relative/path/to/initialValuesFile<br>
    simulation:             relative/path/to/simulationFile<br>
    proposals:              relative/path/to/proposalsFile<br>
    update strategy:        relative/path/to/updateFile<br>
    to output:              relative/path/to/toOutputFile<br>
    last values:            relative/path/to/lastValuesFile<br>
     <i>(end of file)</i><br>
     The order of the identifier - file name pairs is arbitrary.<br>
     The file may contain comments beginning with a hash '#'<br>
     */
    public static void main(String[] args) throws SyntaxException, IOException{
        
        model = null;
        String userDistFileName = null;
        String modelFileName = null;
        String initialValueFileName = null;
        String simulationFileName = null;
        String proposalFileName = null;
        String updateFileName = null;
        String toOutputFileName = null;
        String lastValuesFileName = null;
        if (args.length != 8 && args.length != 1) {
            System.out.println("Wrong number of command-line arguments. Expected: 1 or 8: either one file with the " +
                    "filenames (user_dist, model, initial values, simulation, proposal, update, to output, last values) " +
                    "or these filenames directly.");
            System.exit(1);
        }
        if (args.length == 1){
            Scanner reader = null;
            int linenum = 0;
            try{
                reader = new Scanner(new File(args[0]));
            } catch (FileNotFoundException fnf){
                fnf.printStackTrace();
                System.exit(1);
            }
            while (reader.hasNextLine()){
                String line = reader.nextLine().trim();
                linenum++;
                if (line.length() <= 0 || line.charAt(0) == '#')
                    continue;
                
                if (line.startsWith("user distributions:")){
                    line = line.substring(19).trim();
                    userDistFileName = line;
                } else if (line.startsWith("model:")){
                    line = line.substring(6).trim();
                    modelFileName = line;
                } else if (line.startsWith("initial values:")){
                    line = line.substring(15).trim();
                    initialValueFileName = line;
                } else if (line.startsWith("simulation:")){
                    line = line.substring(11).trim();
                    simulationFileName = line;
                } else if (line.startsWith("proposals:")){
                    line = line.substring(10).trim();
                    proposalFileName = line;
                } else if (line.startsWith("update strategy:")){
                    line = line.substring(16).trim();
                    updateFileName = line;                    
                } else if (line.startsWith("to output:")){
                    line = line.substring(10).trim();
                    toOutputFileName = line;
                } else if (line.startsWith("last values:")){
                    line = line.substring(12).trim();
                    lastValuesFileName = line;
                } else{
                throw new SyntaxException("Invalid line in" + args[0] + ": (line "+ linenum + ")");
                }
            }
            
            
        } else {
            userDistFileName = args[0];
            modelFileName = args[1];
            initialValueFileName = args[2];
            simulationFileName = args[3];
            proposalFileName = args[4];
            updateFileName = args[5];
            toOutputFileName = args[6];
            lastValuesFileName = args[7];
        }
         if (userDistFileName == null)
            throw new IOException("Cannot proceed without a good userDistFileName");
         if (modelFileName == null)
            throw new IOException("Cannot proceed without a good modelFileName");
        if (initialValueFileName == null)
            throw new IOException("Cannot proceed without a good initialValueFileName");
        if (simulationFileName == null)
            throw new IOException("Cannot proceed without a good simulationFileName");
        if (proposalFileName == null)
            throw new IOException("Cannot proceed without a good proposalFileName");
        if (updateFileName == null)
            throw new IOException("Cannot proceed without a good updateFileName");
        if (toOutputFileName == null)
            throw new IOException("Cannot proceed without a good toOutputFileName");
        if (lastValuesFileName == null)
            throw new IOException("Cannot proceed without a good lastValuesFileName");
        
        System.out.println("PIANOS using files:");
        System.out.println("------------------------");
        System.out.println("user distributions:\t\""+ userDistFileName +"\"");
        System.out.println("model:\t\t\t\""+ modelFileName +"\"");
        System.out.println("initial values:\t\t\""+ initialValueFileName +"\"");
        System.out.println("simulation:\t\t\""+ simulationFileName +"\"");
        System.out.println("proposals:\t\t\""+ proposalFileName +"\"");
        System.out.println("update strategy:\t\""+ updateFileName +"\"");
        System.out.println("to output:\t\t\""+ toOutputFileName +"\"");
        System.out.println("last values:\t\t\""+ lastValuesFileName +"\"");
        System.out.println("------------------------");
        try{
            writeProgram(userDistFileName, modelFileName, initialValueFileName, simulationFileName,
                    proposalFileName, updateFileName, toOutputFileName, lastValuesFileName);
        } catch (Exception e){
            e.printStackTrace();
            System.exit(1);
        }
    }
    
    private static String setCallParameters(ComputationalModel model){
        String callParameters = "";
        LinkedList<Variable> variables = model.getVariableList();
        LinkedList<Entity> entities = model.getEntityList();
        boolean first = true;
        
        for (Variable variableIterator : variables){
            if(first){
                callParameters = variableIterator.getName();
                first = false;
            } else
                callParameters = callParameters + ", " + variableIterator.getName();
            
        }
        
        for (Entity entityIterator : entities){
            variables = entityIterator.getVariableList();
            for (Variable variableIterator : variables){
                callParameters = callParameters + ", " + variableIterator.getName();
            }
        }
        
        return callParameters;
    }
    
    // introduce parameters
    public static ArrayList<String> generateIntroduction(){
        
        assert model != null;
        
        ArrayList<String> toWrite = new ArrayList<String>();
        
        for (Variable v : model.getVariableList()){
            if (v.isInteger())
                toWrite.add("TYPE(variables_int), INTENT(INOUT) :: " + v.getName());
            else
                toWrite.add("TYPE(variables_real), INTENT(INOUT) :: " + v.getName());
        }
        
        for (Entity e : model.getEntityList()){
            ArrayList<Variable> variables = new ArrayList<Variable>(e.getVariableList());
            for (Variable v : variables){
                if (v.isInteger())
                    toWrite.add("TYPE(variables_int), INTENT(INOUT) :: " + v.getName());
                else
                    toWrite.add("TYPE(variables_real), INTENT(INOUT) :: " + v.getName());
            }
        }
        return toWrite;
    }
}

