package PIANOS.datastructures;
import java.util.LinkedList;


/** A representation for a stochastic or functional variable/parameter in a statistical PIANOS model.
 
 @author Group Linja
 */
public class Variable{

    private String name;
    private String algorithm;
    private String proposalStrategy;

    private Distribution proposal;
    private Distribution distribution;
    private Equation equation;

    private LinkedList<Variable> affects;
    private LinkedList<Variable> depends;
    private Entity belongsTo;

    private int column;
    private int missingValues;
    private int updates;

    private boolean data;
    private boolean functional;
    private boolean spatial;
    private boolean typeInteger;
    private boolean printed;

    /** Constructs a new Variable from scratch, ready to take in field values.
     */
    public Variable (){
        distribution = null;
        printed = false;
        typeInteger = false;
        missingValues = 0;
        column = -1;
        proposalStrategy = null;
        affects = new LinkedList<Variable>();
        depends = new LinkedList<Variable>();
        spatial = false;
    }

    /** 
     @return A list of Variables this variable affects in the model.
     */
    public LinkedList<Variable> getAffectsList (){
        return affects;
    }

 /** 
     Adds a new Variable at the end of the list of Variables that this Variable affects.
     */
    public void addAffected (Variable variable){
        affects.add(variable);
    }

    /** 
     @return a list of Variables that this Variable depends on.
     */
    public LinkedList<Variable> getDependsList (){
        return depends;
    }

    /** 
     Adds a new Variable at the end of the list of Variables that this Variables depends on.
     */
    public void addDependence (Variable variable){
        depends.add(variable);
    }

    /** 
     @return the update algorithm used for this Variable (currently not implemented), defaults to M-H
     */
    public String getAlgorithm (){
        return algorithm;
    }

    /** (currently not implemented) This will set the update algorithm for this Variable. 
     Choices would include Metropolis-Hastings and Gibbs sampling.
     */
    public void setAlgorithm (String algorithm){
        this.algorithm = algorithm;
    }

    /** 
     @return the column number for the variable if it is found in data, or -1 if the variable spans the whole
     data matrix (and as such all the columns)
     */
    public int getColumn (){
        return column;
    }

    /** 
     Sets the data column number for the Variable (See getColumn())
     */
    public void setColumn (int column){
        this.column = column;
    }

    /** 
     @return if this Variable can be found in the data of this model.
     */
    public boolean isData (){
        return data;
    }

    /** 
     @param data set this to true if the Variable is found in data.
     */
    public void setData(boolean data) {
        this.data = data;
    }

    /** 
     @return The Equation used for this Variable if it is functional (see Equation)
     */
    public Equation getEquation (){
        return equation;
    }

    public void setEquation (Equation equation){
        this.equation = equation;
    }

    /** 
     @return The Distribution used for this Variable if it is stochastic (not functional) (see Distribution)
     */
    public Distribution getDistribution (){
        return distribution;
    }

    public void setDistribution (Distribution distribution){
        this.distribution = distribution;
    }

    /** 
     @return true if this Variable is functional.
     */
    public boolean isFunctional (){
        return functional;
    }

    public void setFunctional (boolean functional){
        this.functional = functional;
    }

    /** 
     @return true if this Variable is a functional variable that has an expression 
     that consists of the SUM or COUNT statements and a spatial dependence in them, such as
     COUNT(&x)
     */
    public boolean isSpatial (){
        return spatial;
    }

    public void setSpatial (boolean spatial){
        this.spatial = spatial;
    }

    /** 
     @return The name used in the model for this variable and to be used in the Fortran program.
     */
    public String getName (){
        return name;
    }

    public void setName (String name){
        this.name = name;
    }

    /** 
     Sets the proposal Distribution for this variable (see the manual)
     */
    public void setProposal (Distribution proposal){
        this.proposal = proposal;
    }

    public Distribution getProposal(){
        return proposal;
    }

    /** 
     Sets the update strategy for this variable {sequential, random}
     */
    public void setStrategy (String strategy){
        this.proposalStrategy = strategy;
    }

    public String getStrategy (){
        return proposalStrategy;
    }

        /** 
     Sets the number of updates this variable should have before end of simulation. 
     */
    public void setUpdates (int updates){
        this.updates = updates;
    }

    public int getUpdates (){
        return updates;
    }

    /** 
     @return How many missing values this variable has (if it's a data variable that has missing values),
     otherwise returns 0.
     */
    public int getMissingValueCount (){
        return missingValues;
    }

     /** 
     Adds one to the number of missing values. Used by the Parser.
     */
    public void incrementMissingValues (){
        missingValues++;
    }

    /** 
     @return true if this variable should have integer values.
     */
    public boolean isInteger (){
        return typeInteger;
    }

    public void setType (boolean typeInteger){
        this.typeInteger = typeInteger;
    }

    /** 
     @return The entity this Variable belongs to.
     */
    public Entity getEntity (){
        return belongsTo;
    }

    public void setEntity (Entity belongsTo){
        this.belongsTo = belongsTo;
    }

    /** 
    Sets that this Variable should be printed.
     */
    public void setPrinted (){
        printed = true;
    }

    public boolean isPrinted (){
        return printed;
    }

    /** 
    Runs this Variable through a health check. Returns true if it is alright. Called by the parser at end of reading.
     @return true if this variable is valid.
     */
    public boolean isOk (){
        if (data && missingValues > 0 && proposal == null)
            return false;
        if (data && missingValues > 0 && distribution == null)
            return false;
        if (!functional && !data && proposal == null)
            return false;
        if (functional && equation == null)
            return false;
        if (!functional && !data && distribution == null)
            return false;

        if (belongsTo != null && !belongsTo.isMatrix()){
            for (Variable var : affects){

                if (var.getEntity() == null){
                //One dimensional affects a global
                    System.out.println("one-dimensional affects a global");
                    return false;
                }
                if (!var.getEntity().isMatrix() && var.getEntity() != belongsTo){
                //Affected is one dimensional and the entity is not the same as the one this variable belongs to
                    System.out.println("one-dimensional affects another one-dimensional in a different entity");
                    return false;
                }
                if (var.getEntity().isMatrix())
                //Affected is a matrix
                    if (var.getEntity().getXCoordinate() != belongsTo && var.getEntity().getYCoordinate() != belongsTo){
                    //This variable's entity isn't either X or Y coordinate of the matrix
                    System.out.println("one-dimensional affects an independent intersection");
                        return false;
                }
            }

        }else if (belongsTo != null && belongsTo.isMatrix()){
            for (Variable var : affects){

                if (var.getEntity() == null){
                //Matrix affects a global
                    System.out.println("Two-dim affects a global");
                    return false;
                }
                if (!var.getEntity().isMatrix()){
                //Affected is one dimensional
                    System.out.println("Intersection variable "+ this +" affects the one-dimensional variable "+ var);
                    return false;
                }
                if (var.getEntity().isMatrix() && var.getEntity() != belongsTo){
                //Affected is a matrix and the entity is not the same as the one this variable belongs to
                    System.out.println("two-dim affects an independent intersection");
                    return false;
             }
            }
        }

        //Check that spatial matrix exists
        if (spatial){

            if (belongsTo == null)
            //Global variable can not be spatial
                return false;

            if (!belongsTo.isMatrix() && !isSpatial())
            //Variable is one dimensional, but the Entity doesn't have a spatial matrix
                return false;

            if (belongsTo.isMatrix())
            //Matrix...
                if (!belongsTo.getXCoordinate().isSpatial() && !belongsTo.getYCoordinate().isSpatial())
                //Neither the X or Y cordinate is spatial
                    return false;
        }

        return true;
    }

    /** Prints a limited amount of information about this Variable similar to a model 
     file variable definition line.
     */
    public String toString(){
        String toReturn = "";
        if (this.isInteger())
            toReturn += "INTEGER ";
        else
            toReturn += "REAL ";

        toReturn += this.getName();
        if (this.isData()){
            toReturn += "(";
            if (this.getColumn() != -1)
            toReturn += this.getColumn();
            else
            toReturn += "*";
            toReturn += ")";
        }
        if (this.isFunctional()){
            toReturn += " = ";
            toReturn += this.getEquation();
        }else{
            if (this.getDistribution() != null){
                toReturn += " ~ ";
                toReturn += this.getDistribution();
            }
            else if (this.isData()){
                toReturn += " is plain data";
            }else
                toReturn += " is not data, stochastic or functional ???";
            }

        if (!this.isFunctional()){
            if (this.getProposal() != null){
                toReturn += " proposal distribution: " + this.getProposal();
        }else
            toReturn += " no proposal distribution ";
        }

        if (this.isPrinted())
            toReturn += ", is printed";

        if (this.getUpdates() != 0)
            toReturn += ", updates=" + this.getUpdates();

        if (this.getMissingValueCount() != 0)
            toReturn += ", " + this.getMissingValueCount() + " missing values";

        return toReturn;
    }

}
