package PIANOS.generator;

import PIANOS.datastructures.*;
import PIANOS.io.*;
import PIANOS.*;

import java.util.*;
import java.io.*;

/**
 * This class is used to generate the Fortran module "output".
 */
public class Output {
    public static final String realfmt = "F15.4";
    public static final String integerfmt = "I10";
    /**
     * Generates and writes the Fortran module output.f90. 
     *
     * @param  callParameters The correct parameter order to be used in subroutine headers
     * @param  model The model this simulation is related to
     */
    public static void generateOutput(String callParameters, ComputationalModel model) 
    throws IOException {

        FortranWriter out = new FortranWriter("output.f90");
        ArrayList <String> result = new ArrayList<String>();
        
        result.add("MODULE output");
        result.add("USE definitions");
        result.add("IMPLICIT NONE");
        result.add("");
        result.add("CONTAINS");
        result.addAll(generateWriteOutput(model, callParameters));
        result.add("");
        result.addAll(generateWriteSummary(model, callParameters));
        result.add("");
        result.addAll(generateWriteLastValues(model, callParameters));
        result.add("END MODULE output");
        out.write(result);
	}

    private static ArrayList<String> generateWriteOutput(ComputationalModel model, 
        String callParameters) {

        ArrayList<Variable> allVariables = new ArrayList<Variable>(model.getVariableList());
        for (Entity e : model.getEntityList()) {
            allVariables.addAll(e.getVariableList());
        }

        ArrayList<String> result = new ArrayList<String>();

        result.add("SUBROUTINE write_output(" + callParameters + ")");
        result.add("IMPLICIT NONE");
        result.addAll(Generator.generateIntroduction());
        result.add("INTEGER :: i, j");
        result.add("INTEGER :: iostatus");
        result.add("CHARACTER (LEN=*), PARAMETER:: filename = '" + model.getOutputFileName() + "'");
        result.add("CHARACTER (LEN=30) :: ich, jch, temp");
        result.add("LOGICAL, SAVE :: written = .FALSE.");

        String explanation = "";

        // We assume that the wanted format for REALs is F15.4 and for INTEGERs I10 

        result.add("IF(.NOT. written) THEN");
        result.add("WRITE(15, FMT = *, IOSTAT = iostatus) '# Model: " + model.getModelFileName() + "'");
        result.addAll(generateCouldNotWrite());
        
        if (model.getUpdateStrategy().equals("sequential")) {
            result.add("WRITE(15, FMT = *, IOSTAT = iostatus) '# Update strategy: sequential, iterations: ', " +
                model.getIterations());
            result.addAll(generateCouldNotWrite());
        }
        else {
            result.add("WRITE(15, FMT = *, IOSTAT = iostatus) '# Update strategy: random'");
            result.addAll(generateCouldNotWrite());
        }
        result.add("WRITE(15, FMT='(A)', ADVANCE = 'NO', IOSTAT = iostatus) '# '");
        for (Variable v : allVariables) {
            if (v.isPrinted() == false) continue;
            
            if (v.getEntity() == null) {
                
                //explanation = explanation + v.getName() + " ";
                result.add("WRITE (15, FMT='(A)', ADVANCE = 'NO', IOSTAT = iostatus) '" 
                    + v.getName() + " '");
                result.addAll(generateCouldNotWrite());
            }
            else if (v.getEntity().isMatrix() == false) {
                result.add("DO i = 1, " + v.getEntity().getSize());
                result.add("WRITE (ich, FMT='(I10)') i");
                result.add("ich = ADJUSTL(ich)");
                result.add("WRITE (15, FMT='(A)', ADVANCE = 'NO', IOSTAT = iostatus) '"+ v.getName() +"_'");
                result.add("WRITE (15, FMT='(A)', ADVANCE = 'NO', IOSTAT = iostatus) TRIM(ich)");
                result.add("WRITE (15, FMT='(A)', ADVANCE = 'NO', IOSTAT = iostatus) ' '");
                result.addAll(generateCouldNotWrite());
                result.add("END DO");

            }
            else {

                result.add("DO j = 1, " + v.getEntity().getXCoordinate().getSize());
                result.add("DO i = 1, " + v.getEntity().getYCoordinate().getSize());
                result.add("WRITE (ich, FMT='(I10)') i");
                result.add("ich = ADJUSTL(ich)");
                result.add("WRITE (jch, FMT='(I10)') j");
                result.add("jch = ADJUSTL(jch)");
                
                result.add("WRITE (15, FMT='(A)', ADVANCE = 'NO', IOSTAT = iostatus) '"+ v.getName() +"_'");
                result.add("WRITE (15, FMT='(A)', ADVANCE = 'NO', IOSTAT = iostatus) TRIM(ich)");
                result.add("WRITE (15, FMT='(A)', ADVANCE = 'NO', IOSTAT = iostatus) '_'");
                result.add("WRITE (15, FMT='(A)', ADVANCE = 'NO', IOSTAT = iostatus) TRIM(jch)");
                result.add("WRITE (15, FMT='(A)', ADVANCE = 'NO', IOSTAT = iostatus) ' '");
                result.addAll(generateCouldNotWrite());
                result.add("END DO");
                result.add("END DO");
            }
        }
        // to get the "advance = no" off
        result.add("WRITE(15, FMT = *, IOSTAT = iostatus) ' '");
        result.addAll(generateCouldNotWrite());
        
        // This is what we want: WRITE(15, FMT = *, IOSTAT = iostatus) 
        // '# alpha x_1_1 x_2_1 x_3_1 x_4_1 x_5_1 ... '
        // result.add("WRITE(15, FMT = *, IOSTAT = iostatus) '# " + explanation + "'");
        // result.addAll(generateCouldNotWrite());

        result.add("written = .TRUE.");
        result.add("END IF");

        // This is what we want: WRITE(15, FMT='(F6.2, 84(I2))', IOSTAT = iostatus) 
        // alpha % one_dim(1) % value, x % two_dim(:, :) % value
        
           for (Variable v : allVariables) {
            if (v.isPrinted() == false) continue;
            
            String fmt = null;
            
            if (v.isInteger())
                fmt = integerfmt;
            else
                fmt = realfmt;
            
            if (v.getEntity() == null) {
                result.add("WRITE (temp, FMT='("+ fmt +")') "+ v.getName() +" % one_dim(1) % value");
                result.add("temp = ADJUSTL(temp)");
                result.add("WRITE (15, FMT='(A)', ADVANCE = 'NO', IOSTAT = iostatus) TRIM(temp)");
                result.add("WRITE (15, FMT='(A)', ADVANCE = 'NO', IOSTAT = iostatus) ' '");
                result.addAll(generateCouldNotWrite());
            }
            else if (v.getEntity().isMatrix() == false) {
                result.add("DO i = 1, " + v.getEntity().getSize());
                result.add("WRITE (temp, FMT='("+ fmt +")') "+ v.getName() +" % one_dim(i) % value");                
                result.add("temp = ADJUSTL(temp)");
                result.add("WRITE (15, FMT='(A)', ADVANCE = 'NO', IOSTAT = iostatus) TRIM(temp)");
                result.add("WRITE (15, FMT='(A)', ADVANCE = 'NO', IOSTAT = iostatus) ' '");
                result.addAll(generateCouldNotWrite());
                result.add("END DO");
            }
            else {
               result.add("DO j = 1, " + v.getEntity().getXCoordinate().getSize());
                result.add("DO i = 1, " + v.getEntity().getYCoordinate().getSize());
                result.add("WRITE (temp, FMT='("+ fmt +")') "+ v.getName() +" % two_dim(i, j) % value");                
                result.add("temp = ADJUSTL(temp)");
                result.add("WRITE (15, FMT='(A)', ADVANCE = 'NO', IOSTAT = iostatus) TRIM(temp)");
                result.add("WRITE (15, FMT='(A)', ADVANCE = 'NO', IOSTAT = iostatus) ' '");
                result.addAll(generateCouldNotWrite());                
                result.add("END DO");
                result.add("END DO");
            }
            
        }
        result.add("WRITE (15, FMT='(A)', ADVANCE = 'YES', IOSTAT = iostatus) ' '");
        result.addAll(generateCouldNotWrite());

        result.add("END SUBROUTINE write_output");
	    return result;
    }

    
    private static ArrayList<String> generateWriteSummary(ComputationalModel model, 
        String callParameters) {

        ArrayList<Variable> allVariables = new ArrayList<Variable>(model.getVariableList());
        for (Entity e : model.getEntityList()) {
            allVariables.addAll(e.getVariableList());
        }

        ArrayList<String> result = new ArrayList<String>();

        result.add("SUBROUTINE write_summary(" + callParameters + ")");
        result.add("IMPLICIT NONE");
        result.addAll(Generator.generateIntroduction());
        result.add("INTEGER :: iostatus, i, j");
        result.add("CHARACTER(LEN=*), PARAMETER :: filename = '"+ model.getSummaryFileName() 
            + "'");        
        result.add("CHARACTER(LEN=30) :: ich, jch, temp, tempcount");
        result.add("OPEN(UNIT=21, FILE=filename, IOSTAT = iostatus)");
        result.addAll(generateCouldNotOpen());

        result.add("WRITE(21, '(A)', IOSTAT = iostatus) '?? simulation summary'");
        result.addAll(generateCouldNotWrite());

        for (Variable variable : allVariables) {

            if (variable.isPrinted() == false) continue;
  
            String name = variable.getName();
            result.add("WRITE(21, '(A)', IOSTAT = iostatus) '? " + name + "'");
            result.addAll(generateCouldNotWrite());
        
            if (model.getUpdateStrategy().equals("sequential")) {

                int iterations = model.getIterations();
                if (variable.getEntity() == null) {
                    result.add("WRITE(temp, FMT = '("+realfmt+")', IOSTAT = iostatus) REAL(" + 
                        name +" % one_dim(1) % successful_changes)/" + iterations);
                    result.add("temp = ADJUSTL(temp)");
                    result.add("WRITE(21, FMT = '(A)', ADVANCE ='NO', IOSTAT = iostatus) '" + iterations +"'");                    
                    result.add("WRITE(21, FMT = '(A)', ADVANCE ='NO', IOSTAT = iostatus) ' '");
                    result.add("WRITE(21, FMT = '(A)', ADVANCE ='YES', IOSTAT = iostatus) TRIM(temp)");
                    result.addAll(generateCouldNotWrite());
                }
                else if (variable.getEntity().isMatrix() == false) {
                    result.add("DO i=1, " + variable.getEntity().getSize());
                    result.add("WRITE (ich, FMT='(I10)') i");
                    result.add("ich = ADJUSTL(ich)");
                    result.add("WRITE(21, FMT = '(A)', ADVANCE='NO', IOSTAT = iostatus) '? " + name + "_'");
                    result.add("WRITE(21, FMT = '(A)', ADVANCE='YES', IOSTAT = iostatus) TRIM(ich)");
                    result.addAll(generateCouldNotWrite());
                    
                    result.add("WRITE(temp, FMT = '("+realfmt+")', IOSTAT = iostatus) REAL(" + 
                        name +" % one_dim(i) % successful_changes)/" + iterations);
                    result.add("temp = ADJUSTL(temp)");
                    result.add("WRITE(21, FMT = '(A)', ADVANCE ='NO', IOSTAT = iostatus) '" + iterations +"'");                    
                    result.add("WRITE(21, FMT = '(A)', ADVANCE ='NO', IOSTAT = iostatus) ' '");
                    result.add("WRITE(21, FMT = '(A)', ADVANCE ='YES', IOSTAT = iostatus) TRIM(temp)");
                    result.addAll(generateCouldNotWrite());
                    result.add("END DO");
                }
                else {
                    result.add("DO j=1, " + variable.getEntity().getXCoordinate().getSize());
                    result.add("DO i=1, " + variable.getEntity().getYCoordinate().getSize());
                    result.add("WRITE (ich, FMT='(I10)') i");
                    result.add("ich = ADJUSTL(ich)");
                    result.add("WRITE (jch, FMT='(I10)') j");
                    result.add("jch = ADJUSTL(jch)");
                    result.add("WRITE(21, FMT = '(A)', ADVANCE='NO', IOSTAT = iostatus) '? " + name + "_'");
                    result.add("WRITE(21, FMT = '(A)', ADVANCE='NO', IOSTAT = iostatus) TRIM(ich)");
                    result.add("WRITE(21, FMT = '(A)', ADVANCE='NO', IOSTAT = iostatus) '_'");
                    result.add("WRITE(21, FMT = '(A)', ADVANCE='YES', IOSTAT = iostatus) TRIM(jch)");
                    result.addAll(generateCouldNotWrite());
                    result.add("WRITE(temp, FMT = '("+realfmt+")', IOSTAT = iostatus) REAL(" + 
                        name +" % two_dim(i, j) % successful_changes)/" + iterations);
                    result.add("temp = ADJUSTL(temp)");
                    result.add("WRITE(21, FMT = '(A)', ADVANCE ='NO', IOSTAT = iostatus) '" + iterations +"'");                    
                    result.add("WRITE(21, FMT = '(A)', ADVANCE ='NO', IOSTAT = iostatus) ' '");
                    result.add("WRITE(21, FMT = '(A)', ADVANCE ='YES', IOSTAT = iostatus) TRIM(temp)");
                    result.addAll(generateCouldNotWrite());
                    result.add("END DO");
                    result.add("END DO");
                }
            }
            else {
                            
                if (variable.getEntity() == null) {
                    result.add("WRITE(tempcount, FMT = '(I10)', IOSTAT = iostatus) " + name + " % one_dim(1) % update_count");
                    result.add("tempcount = ADJUSTL(tempcount)");
                    result.add("WRITE(temp, FMT = '("+realfmt+")', IOSTAT = iostatus) REAL(" + 
                        name +" % one_dim(1) % successful_changes)/ " + name + " % one_dim(1) % update_count");
                    result.add("temp = ADJUSTL(temp)");
                    result.add("WRITE(21, FMT = '(A)', ADVANCE ='NO', IOSTAT = iostatus) TRIM(tempcount)");
                    result.add("WRITE(21, FMT = '(A)', ADVANCE ='NO', IOSTAT = iostatus) ' '");
                    result.add("WRITE(21, FMT = '(A)', ADVANCE ='YES', IOSTAT = iostatus) TRIM(temp)");
                    result.addAll(generateCouldNotWrite());
                }
                else if (variable.getEntity().isMatrix() == false) {
                    result.add("DO i=1, " + variable.getEntity().getSize());
                    result.add("WRITE (ich, FMT='(I10)') i");
                    result.add("ich = ADJUSTL(ich)");
                    result.add("WRITE(21, FMT = '(A)', ADVANCE='NO', IOSTAT = iostatus) '? " + name + "_'");
                    result.add("WRITE(21, FMT = '(A)', ADVANCE='YES', IOSTAT = iostatus) TRIM(ich)");
                    result.addAll(generateCouldNotWrite());
                    
                    result.add("WRITE(tempcount, FMT = '(I10)', IOSTAT = iostatus) " + name + " % one_dim(i) % update_count");
                    result.add("tempcount = ADJUSTL(tempcount)");
                    result.add("WRITE(temp, FMT = '("+realfmt+")', IOSTAT = iostatus) REAL(" + 
                        name +" % one_dim(i) % successful_changes)/ " + name + " % one_dim(i) % update_count");
                    result.add("temp = ADJUSTL(temp)");
                    result.add("WRITE(21, FMT = '(A)', ADVANCE ='NO', IOSTAT = iostatus) TRIM(tempcount)");
                    result.add("WRITE(21, FMT = '(A)', ADVANCE ='NO', IOSTAT = iostatus) ' '");
                    result.add("WRITE(21, FMT = '(A)', ADVANCE ='YES', IOSTAT = iostatus) TRIM(temp)");
                    result.addAll(generateCouldNotWrite());
                    result.add("END DO");
                }
                else {
                    result.add("DO j=1, " + variable.getEntity().getXCoordinate().getSize());
                    result.add("DO i=1, " + variable.getEntity().getYCoordinate().getSize());
                    result.add("WRITE (ich, FMT='(I10)') i");
                    result.add("ich = ADJUSTL(ich)");
                    result.add("WRITE (jch, FMT='(I10)') j");
                    result.add("jch = ADJUSTL(jch)");
                    result.add("WRITE(21, FMT = '(A)', ADVANCE='NO', IOSTAT = iostatus) '? " + name + "_'");
                    result.add("WRITE(21, FMT = '(A)', ADVANCE='NO', IOSTAT = iostatus) TRIM(ich)");
                    result.add("WRITE(21, FMT = '(A)', ADVANCE='NO', IOSTAT = iostatus) '_'");
                    result.add("WRITE(21, FMT = '(A)', ADVANCE='YES', IOSTAT = iostatus) TRIM(jch)");
                    result.addAll(generateCouldNotWrite());
                    
                    result.add("WRITE(tempcount, FMT = '(I10)', IOSTAT = iostatus) " + name + " % two_dim(i, j) % update_count");
                    result.add("tempcount = ADJUSTL(tempcount)");
                    result.add("WRITE(temp, FMT = '("+realfmt+")', IOSTAT = iostatus) REAL(" + 
                        name +" % two_dim(i, j) % successful_changes)/ " + name + " % two_dim(i, j) % update_count");
                    result.add("temp = ADJUSTL(temp)");
                    result.add("WRITE(21, FMT = '(A)', ADVANCE ='NO', IOSTAT = iostatus) TRIM(tempcount)");
                    result.add("WRITE(21, FMT = '(A)', ADVANCE ='NO', IOSTAT = iostatus) ' '");
                    result.add("WRITE(21, FMT = '(A)', ADVANCE ='YES', IOSTAT = iostatus) TRIM(temp)");

                    result.addAll(generateCouldNotWrite());
                    result.add("END DO");
                    result.add("END DO");
                }
            }    
        }

        result.add("CLOSE(21)");
        result.add("END SUBROUTINE write_summary");

        return result;
    }

    private static ArrayList<String> generateCouldNotWrite() {
        
        ArrayList<String> result = new ArrayList<String>();
        result.add("IF (iostatus /= 0) THEN");
        result.add("WRITE (*, *) 'Could not write to ', filename, '. Exiting program.'");
        result.add("STOP");
        result.add("END IF");
        return result;
    }

    private static ArrayList<String> generateCouldNotOpen() {
        
        ArrayList<String> result = new ArrayList<String>();
        result.add("IF (iostatus /= 0) THEN");
        result.add("WRITE (*, *) 'Could not open ', filename, '. Exiting program.'");
        result.add("STOP");
        result.add("END IF");
        return result;
    }

    private static ArrayList<String> generateWriteLastValues(ComputationalModel model, String callParameters) {
        // the string parameters tells the correct order of the parameters
        // the initializations can be generated based on Variables.
        ArrayList<Variable> allVariables = new ArrayList<Variable>(model.getVariableList());
        for (Entity e : model.getEntityList()) {
            allVariables.addAll(e.getVariableList());
        }
        
        ArrayList<String> result = new ArrayList<String>();

        result.add("SUBROUTINE write_last_values(" + callParameters + ")");
        result.add("IMPLICIT NONE");
        
        result.addAll(Generator.generateIntroduction());

        result.add("INTEGER :: iostatus, i, j");
        result.add("CHARACTER (LEN=*), PARAMETER :: filename = '" + model.getLastValuesFileName() + "'"); 
        result.add("CHARACTER (LEN=30) :: temp");
        result.add("OPEN(UNIT=20, FILE=filename, IOSTAT = iostatus)");
        result.addAll(generateCouldNotOpen());
        
        for (Variable variable : allVariables) {
            
            //System.out.println("GENERATOR DEBUG: Variable name : " + variable.getName() 
            //    + ", functional: " + variable.isFunctional());
            
            if (variable.isFunctional()) continue;
            if (variable.isData() && variable.getMissingValueCount() == 0) continue;
            String fmt = null;
            if (variable.isInteger())
                fmt = integerfmt;
            else
                fmt = realfmt;
            
            String name = variable.getName();
            if (variable.getEntity() == null) {
                result.add("WRITE(20, '(A)', IOSTAT = iostatus) '? " + name + "'");
                result.addAll(generateCouldNotWrite());
                result.add("WRITE(temp, FMT='("+ fmt +")', IOSTAT = iostatus)" + name +" % one_dim(1) % value");
                result.add("temp = ADJUSTL(temp)");
                result.add("WRITE(20, '(A)', IOSTAT = iostatus) TRIM(temp)");
                result.addAll(generateCouldNotWrite());
            }
            else if (variable.getEntity().isMatrix() == false) {
                int size = variable.getEntity().getSize();
                result.add("WRITE(20, '(A)', IOSTAT = iostatus) '? " + name + " 1: " + 
                    size + "'");
                result.addAll(generateCouldNotWrite());
                result.add("DO i = 1, " + size);
                result.add("WRITE(temp, FMT='("+ fmt +")', IOSTAT = iostatus)" + name +" % one_dim(i) % value");
                result.add("temp = ADJUSTL(temp)");
                result.add("WRITE(20, '(A)', IOSTAT = iostatus) TRIM(temp)");
                result.addAll(generateCouldNotWrite());
                result.add("END DO");
            }
            else {
                int ysize = variable.getEntity().getYCoordinate().getSize();
                int xsize = variable.getEntity().getXCoordinate().getSize();
                result.add("WRITE(20, '(A)', IOSTAT = iostatus) '? " + name + " 1:" + ysize + " 1:" + xsize + "'");
                result.addAll(generateCouldNotWrite());

                result.add("DO i = 1, " + ysize);
                result.add("DO j = 1, " + xsize);
                result.add("WRITE(temp, FMT='("+ fmt +")', IOSTAT = iostatus)" + name +" % two_dim(i, j) % value");
                result.add("temp = ADJUSTL(temp)");
                result.add("WRITE(20, '(A)', IOSTAT = iostatus, ADVANCE='NO') TRIM(temp)");
                result.add("WRITE(20, '(A)', IOSTAT = iostatus, ADVANCE='NO') ' '");
                result.addAll(generateCouldNotWrite());
                result.add("END DO");
                result.add("WRITE(20, '(A)', IOSTAT = iostatus, ADVANCE='YES') ' '");
                result.add("END DO");
            }
        }

        result.add("CLOSE(20)");
        result.add("END SUBROUTINE write_last_values");
        return result;
    }
    
}
