package PIANOS.io;
import PIANOS.datastructures.*;
import PIANOS.exceptions.*;

// data structures
import java.util.HashMap;
import java.util.LinkedList;
import java.util.ArrayList;

// general reading
import java.util.Scanner;

// reading the data files
import java.util.StringTokenizer;

// parsing the equations correctly
import java.util.regex.Pattern;
import java.util.regex.Matcher;

// File, IOException, etc...
import java.io.*;

/**
 * The PIANOS parser
 * Reads input files with readModel and returns the corresponding model.
 * @author Eemil Lagerspetz
 */
public class ComputationalModelParser{
    
    private ComputationalModelParser(){
    }
    
    private static int line;
    /** Public for testing purposes, do not modify.
     */
    public static ArrayList<Variable> toBeChecked;
    /** Public for testing purposes, do not modify.
     */
    public static HashMap<Variable, String> spatAffectedMap;
    private static HashMap<String, Entity> belongMap;
    
    public static ComputationalModel modelTestOnly(String modelFileName,
            String proposalFileName,
            DistributionFactory factory)
            throws IOException, SyntaxException, MissingDistributionException{
        
        // we are supposed to create a ComputationalModel, so let's do just that.
        // First initialize the files...
        if (modelFileName == null)
            throw new IOException("Can't read a file with null name as modelFileName");
        if (proposalFileName == null)
            throw new IOException("Can't read a file with null name as proposalFileName");
        
        File modelFile = new File(modelFileName);
        File proposalFile = new File(proposalFileName);
        
        // build the components...
        int iterations = 0;
        int burnIn = 0;
        int thinning = 0;
        int neighbours = 0;
        String updateStrategy = "";
        
        LinkedList<Variable> variableList = new LinkedList<Variable>();
        LinkedList<Entity> entityList = new LinkedList<Entity>();
        HashMap<String, Entity> entityMapper = new HashMap<String, Entity>();
        HashMap<String, Variable> variableMapper = new HashMap<String, Variable>();
        
        // and start setting them:
        toBeChecked = new ArrayList<Variable>();
        spatAffectedMap = new HashMap<Variable, String>();
        
        readModelFile(modelFile, variableList, entityList, entityMapper, variableMapper, factory);
        
        // now we might have stuff in toBeChecked and spatAffectedMap.
        
        /* These operations still need to be done for stuff in them.
           toAdd.addDependence(dep);
           dep.addAffected(toAdd);
           variables.add(dep);
         */
        
        checkSpats(variableMapper, modelFile);
        
        // now they should be alright.
        
        readProposal(proposalFile, variableMapper, factory);
        
        readEntityData(entityList);
        
        neighbours = countNeighbours(entityList);
        
        checkVariablesAreOk(new ArrayList<Variable>(variableMapper.values()));
        
        ComputationalModel model = new ComputationalModel(iterations, burnIn, thinning, neighbours,
                updateStrategy, variableList, entityList, entityMapper, variableMapper, modelFileName,
                null, null, null);
        return model;
    }
    
    public static ComputationalModel testNoInitialValues(String modelFileName,
            String simulationFileName,
            String proposalFileName,
            String updateFileName,
            String toOutputFileName, DistributionFactory factory)
            throws IOException, SyntaxException, MissingDistributionException{
        if (modelFileName == null)
            throw new IOException("Can't read a file with null name as modelFileName");
        //	if (initialValueFileName == null)
        //	    throw new IOException("Can't read a file with null name as initialValueFileName");
        if (simulationFileName == null)
            throw new IOException("Can't read a file with null name as simulationFileName");
        if (proposalFileName == null)
            throw new IOException("Can't read a file with null name as proposalFileName");
        if (updateFileName == null)
            throw new IOException("Can't read a file with null name as updateFileName");
        if (toOutputFileName == null)
            throw new IOException("Can't read a file with null name as toOutputFileName");
        
        File modelFile = new File(modelFileName);
        //	File initialValuesFile = new File(initialValueFileName);
        File simulationFile = new File(simulationFileName);
        File proposalFile = new File(proposalFileName);
        File updateFile = new File(updateFileName);
        File toOutputFile = new File(toOutputFileName);
        
        // build the components...
        int iterations = 0;
        int burnIn = 0;
        int thinning = 0;
        int neighbours = 0;
        String updateStrategy = "";
        
        LinkedList<Variable> variableList = new LinkedList<Variable>();
        LinkedList<Entity> entityList = new LinkedList<Entity>();
        HashMap<String, Entity> entityMapper = new HashMap<String, Entity>();
        HashMap<String, Variable> variableMapper = new HashMap<String, Variable>();
        
        // and start setting them:
        toBeChecked = new ArrayList<Variable>();
        spatAffectedMap = new HashMap<Variable, String>();
        
        readModelFile(modelFile, variableList, entityList, entityMapper, variableMapper, factory);
        
        // now we might have stuff in toBeChecked and spatAffectedMap.
        
        /* These operations still need to be done for stuff in them.
           toAdd.addDependence(dep);
           dep.addAffected(toAdd);
           variables.add(dep);
         */
        
        checkSpats(variableMapper, modelFile);
        
        // now they should be alright.
        
        int[] temp = readSimulation(simulationFile);
        burnIn = temp[0];
        thinning = temp[1];
        
        Object[] updateTemp = readUpdate(updateFile, variableMapper , updateStrategy);
        updateStrategy = (String)updateTemp[0];
        iterations = (Integer)updateTemp[1];
        
        String[] opFiles = readOutput(toOutputFile, variableMapper);
        
        readProposal(proposalFile, variableMapper, factory);
        
        readEntityData(entityList);
        
        neighbours = countNeighbours(entityList);
        //	readInitialValues(initialValuesFile, variableMapper);
        
        checkVariablesAreOk(new ArrayList<Variable>(variableMapper.values()));
        
        ComputationalModel model = new ComputationalModel(iterations, burnIn, thinning, neighbours,
                updateStrategy, variableList, entityList, entityMapper, variableMapper, modelFileName,
                null, opFiles[0], opFiles[1]);
        return model;
    }
    
    
    /**
     * The main interface for the Parser. All filenames are valid with a relative path.
     * @param modelFileName the name of the model file
     * @param initialValueFileName the name of the initial
     * values file
     * @param simulationFileName the name of the model file
     * @param proposalFileName the name of the proposal distributions
     * file
     * @param updateFileName the name of the variables update strategy
     * file
     * @param toOutputFileName the name of the to_output file
     * @param factory the DistributionFactory with the user distributions used
     @return the complete model data structure constructed from the files given.
     */
    
    public static ComputationalModel readModel(String modelFileName,
            String initialValueFileName,
            String simulationFileName,
            String proposalFileName,
            String updateFileName,
            String toOutputFileName, DistributionFactory factory)
            throws IOException, SyntaxException, MissingDistributionException{
        
        // we are supposed to create a ComputationalModel, so let's do just that.
        // First initialize the files...
        if (modelFileName == null)
            throw new IOException("Can't read a file with null name as modelFileName");
        if (initialValueFileName == null)
            throw new IOException("Can't read a file with null name as initialValueFileName");
        if (simulationFileName == null)
            throw new IOException("Can't read a file with null name as simulationFileName");
        if (proposalFileName == null)
            throw new IOException("Can't read a file with null name as proposalFileName");
        if (updateFileName == null)
            throw new IOException("Can't read a file with null name as updateFileName");
        if (toOutputFileName == null)
            throw new IOException("Can't read a file with null name as toOutputFileName");
        
        File modelFile = new File(modelFileName);
        File initialValuesFile = new File(initialValueFileName);
        File simulationFile = new File(simulationFileName);
        File proposalFile = new File(proposalFileName);
        File updateFile = new File(updateFileName);
        File toOutputFile = new File(toOutputFileName);
        
        // build the components...
        int iterations = 0;
        int burnIn = 0;
        int thinning = 0;
        int neighbours = 0;
        String updateStrategy = "";
        
        LinkedList<Variable> variableList = new LinkedList<Variable>();
        LinkedList<Entity> entityList = new LinkedList<Entity>();
        HashMap<String, Entity> entityMapper = new HashMap<String, Entity>();
        HashMap<String, Variable> variableMapper = new HashMap<String, Variable>();
        
        // and start setting them:
        toBeChecked = new ArrayList<Variable>();
        spatAffectedMap = new HashMap<Variable, String>();
        
        
        System.out.print("Reading the model file...");
        readModelFile(modelFile, variableList, entityList, entityMapper, variableMapper, factory);
        System.out.println("\t\t\t\t\t[OK]");
        // now we might have stuff in toBeChecked and spatAffectedMap.
        
        /* These operations still need to be done for stuff in them.
           toAdd.addDependence(dep);
           dep.addAffected(toAdd);
           variables.add(dep);
         */
        
        checkSpats(variableMapper, modelFile);
        
        
        // now they should be alright.
        
        System.out.print("Reading simulation parameters...");
        int[] temp = readSimulation(simulationFile);
        burnIn = temp[0];
        thinning = temp[1];
        
        System.out.println("\t\t\t\t[OK]");
        
        System.out.print("Reading update file...");
        Object[] updateTemp = readUpdate(updateFile, variableMapper , updateStrategy);
        updateStrategy = (String)updateTemp[0];
        iterations = (Integer)updateTemp[1];
        System.out.println("\t\t\t\t\t\t[OK]");
        
        System.out.print("Getting output variable names...");        
        String[] opFiles = readOutput(toOutputFile, variableMapper);
        
        System.out.println("\t\t\t\t[OK]");
        
        
        System.out.print("Validating proposal distributions...");
        readProposal(proposalFile, variableMapper, factory);
        System.out.println("\t\t\t\t[OK]");
        
        
        System.out.print("Validating data...");
        readEntityData(entityList);
        System.out.println("\t\t\t\t\t\t[OK]");
        
        System.out.print("Finding maximum number of neighbours...");
        neighbours = countNeighbours(entityList);
        System.out.println("\t\t\t\t[OK]");
        
        System.out.print("Checking initial values...");
        readInitialValues(initialValuesFile, variableMapper);
        System.out.println("\t\t\t\t\t[OK]");
        
        System.out.print("Model check...");
        checkVariablesAreOk(new ArrayList<Variable>(variableMapper.values()));
        System.out.println("\t\t\t\t\t\t\t[OK]");
        
        ComputationalModel model = new ComputationalModel(iterations, burnIn, thinning, neighbours,
                updateStrategy, variableList, entityList, entityMapper, variableMapper, modelFileName,
                initialValueFileName, opFiles[0], opFiles[1]);
        
        System.out.println("Model built.");
        return model;
    }
    
    private static void checkSpats(HashMap <String, Variable> variableMapper, File modelFile)
    throws SyntaxException{
        for (Variable aff : toBeChecked){
            String varname = spatAffectedMap.get(aff);
            Variable var = variableMapper.get(varname);
            if (aff == null)
                throw new SyntaxException("In "+ modelFile + ": Variable "+ varname +
                        " is used in a COUNT or SUM statement but never defined");
            aff.addDependence(var);
            var.addAffected(aff);
            // It's good that we defined that SUM and COUNT must appear alone.
            // otherwise this would have been VERY troublesome.
            aff.getEquation().setParameters(var);
        }
    }
    
    /** Reads the model file and puts the information into the maps and lists provided.
     *
     */
    
    private static void readModelFile(File file, LinkedList<Variable> variableList,
            LinkedList<Entity> entityList, HashMap<String, Entity> entityMapper,
            HashMap<String, Variable> variableMapper, DistributionFactory fact)
            throws IOException, SyntaxException, MissingDistributionException{
        // reads the model, creates data structures from it, sets the variables' sizes correctly
        // with help from readData(which creates the missing binary matrix and the modified data
        // file as a side-product)
        
        
        Scanner reader = new Scanner(file);
        String temp = null;
        line = 0;
        while (reader.hasNextLine()){
            temp = reader.nextLine();
            line++;
            temp = temp.trim();
            
            if (temp.length() == 0 || temp.charAt(0) == '#')
                continue;
            
            // now we have a non-comment non-zero line.
            
            // it can either be a global variable starting with
            // REAL or INTEGER,
            // or it can be an ENTITY definition line starting with ENTITY.
            
            if (temp.startsWith("REAL")){
                temp = temp.substring(4).trim();
                readglobal(file, false, temp, variableMapper, variableList, fact);
            } else if (temp.startsWith("INTEGER")){
                temp = temp.substring(7).trim();
                readglobal(file, true, temp, variableMapper, variableList, fact);
            } else if (temp.startsWith("ENTITY")){
                temp = temp.substring(6).trim();
                readEntity(file, entityList, entityMapper, variableMapper, temp, reader, fact);
            } else
                throw new SyntaxException("No entity or variable on line " + line);
            
        }
        reader.close();
    }
    
    /** Reads a global variable's information from the line provided and saves it to the variablelist and mapper.
     *
     */
    
    private static void readglobal(File file, boolean isInteger, String toParse,
            HashMap<String, Variable> variableMapper, LinkedList<Variable> variableList,
            DistributionFactory fact) throws SyntaxException,
            MissingDistributionException {
        Variable toAdd = parseVariable(file, isInteger, toParse, variableMapper, fact);
        toAdd.setEntity(null); // global
        variableMapper.put(toAdd.getName(), toAdd);
        variableList.add(toAdd); // global variables list
    }
    
    /** Reads a variable definition line and returns a representation of the variable, public for testing purposes. Do not call directly.
     *
     */
    
    public static Variable parseVariable(File file, boolean isInteger, String toParse,
            HashMap<String, Variable> variableMapper, DistributionFactory fact) throws
            SyntaxException, MissingDistributionException{
        Variable toAdd = new Variable();
        
        toAdd.setType(isInteger);
        toParse = toParse.replaceAll("\\p{Space}+", "");
        /*
         *Now we have either
         *nam4e~dist(par,par2,..,par83)
         *or
         *nam4e=expression
         *or
         *nam4e(index)
         *or
         *nam4e(index)~dist(par,...)
         *or
         *nam4e(index)=expr
         */
        
        int parenIndex = toParse.indexOf('(');
        int eqIndex = toParse.indexOf('=');
        int tildeIndex = toParse.indexOf('~');
        if (parenIndex == -1)
            parenIndex = toParse.length();
        if (eqIndex == -1)
            eqIndex = toParse.length();
        if (tildeIndex == -1)
            tildeIndex = toParse.length();
        
        String name = toParse.substring(0, Math.min(parenIndex,
                Math.min(eqIndex, tildeIndex)));
        
        
        Pattern p = Pattern.compile("[a-z][a-z0-9]*");
        Matcher m = p.matcher(name);
        if (!m.matches())
            throw new SyntaxException("In "+ file + ": Invalid variable name " +
                    "on line " + line);
        
        toAdd.setName(name);
        
        toParse = toParse.substring(name.length()); // the char after the last of name
        
        /*
         *Now we have either
         *~dist(par,par2,..,par83)
         *or
         *=expression
         *or
         *(index)
         *or
         *(index)~dist(par,...)
         *or
         *(index)=expr
         */
        
        if (toParse.length() <= 0)
            throw new SyntaxException("In "+ file +": No column definition, expression or distribution " +
                    "found for variable on line " + line);
        
        if (toParse.charAt(0) == '('){
            // parse the column definition and parse it out
            String col = toParse.substring(1, toParse.indexOf(')'));
            toParse = toParse.substring(toParse.indexOf(')') +1);
            toAdd.setData(true);
            if (col.equals("*")) // *, nothing else
                toAdd.setColumn(-1);
            else{
                try{
                    int temp = Integer.parseInt(col);
                    if (temp < 1)
                        throw new SyntaxException("In "+ file +": Invalid column number on line " + line
                                + " (columns start at 1)");
                    toAdd.setColumn(temp);
                }catch (NumberFormatException nfe){
                    // not *, not an integer, problem
                    throw new SyntaxException("In "+ file + ": Invalid column number on line " + line);
                }
            }
        } else { // no column def
            toAdd.setData(false);
        }
        
        
        // name, column, isData are set
        
          /*
           *Now we have either
           *~dist(par,par2,..,par83)
           *or
           *=expression
           *or
           *
           */
        
        
        if (toParse.length() <= 0)
            return toAdd;
        
         /*
          *Now we have either
          *~dist(par,par2,..,par83)
          *or
          *=expression
          */
        
        if (toParse.charAt(0) == '~') {
            toAdd.setFunctional(false);
            String distStr = null;
            if (toParse.indexOf('(') < 0)
                distStr = "";
            else
                distStr = toParse.substring(1, toParse.indexOf('(')).trim();
            Distribution dist = fact.getDistribution(distStr);
            // this might throw a MissingDistributionException.
            toAdd.setDistribution(dist);
            
            if (toParse.indexOf(')') < 0){
                throw new SyntaxException("In "+ file +": Mismatching parentheses on line "+ line);
            }
            
            
            toParse = toParse.substring(toParse.indexOf('(') +1, toParse.lastIndexOf(')')).trim();
            // Now we have "param1, param2, ..., paramn"
            
            if (toParse.length() == 0)
                throw new SyntaxException("In "+ file +": Too few parameters on line "+ line);
            
            int index = 0;
            while (toParse != null) {
                int comma = toParse.indexOf(',');
                String param = null;
                if (comma > 0) {
                    param = toParse.substring(0, comma).trim();
                    // there is a comma, so the next parameter should be before it
                    // for the next round:
                    toParse = toParse.substring(comma +1).trim();
                } else {
                    param = toParse.trim();
                    // no comma found, so this is the last parameter
                    
                    // to terminate the loop
                    toParse = null;
                    
                    // are there too few parameters in the file?
                    
                }
                
                // are there too many parameters in the file?
                if (dist.getNumberOfParameters() <= index)
                    throw new SyntaxException("In "+ file +": Too many parameters on line "+ line);
                boolean integer = dist.isInteger(index);
                if (integer){
                    try{
                        int paramInt = Integer.parseInt(param);
                        dist.setParameter(index, paramInt);
                        // it's a good integer, no need to check for anything else
                        
                    } catch(NumberFormatException e){
                        // okay, it's not an Integer.
                        // it may still be a Variable reference
                        // or something malformed
                        p = Pattern.compile("[a-z][a-z0-9]*");
                        
                        m = p.matcher(param);
                        if (!m.matches()){
                            //Ok, it's definitely something malformed.
                            throw new SyntaxException("In "+ file +": Parameter " + (index +1) +
                                    " on line " + line + " is not an integer or a correct variable name");
                        }
                        Variable paramVar = variableMapper.get(param);
                        if (paramVar == null) // run out of tricks
                            throw new SyntaxException("In "+ file +": Parameter " + (index +1) +
                                    " on line " + line + " is not defined before it is used");
                        else{
                            if (!paramVar.isInteger())
                                throw new SyntaxException("In "+ file +": Parameter " + (index +1) +
                                        " on line " + line + " is not of type INTEGER");
                            dist.setParameter(index, paramVar);
                            paramVar.addAffected(toAdd);
                            toAdd.addDependence(paramVar);
                        }
                    }
                } else{
                    // not an integer, so it should be REAL
                    try {
                        double paramDouble = Double.parseDouble(param);
                        dist.setParameter(index, paramDouble);
                        // it's a good double, continue
                        
                    } catch (NumberFormatException nfe){
                        // okay, not a good double.
                        
                        // is it something malformed?
                        p = Pattern.compile("[a-z][a-z0-9]*");
                        m = p.matcher(param);
                        
                        if (!m.matches()){
                            //Ok, it's definitely something malformed.
                            throw new SyntaxException("In "+ file +": Parameter " + (index +1) +
                                    " on line " + line + " is not a decimal number or a correct " +
                                    "variable name");
                        }
                        
                        // is it a Variable?
                        
                        Variable paramVar = variableMapper.get(param);
                        if (paramVar == null) // run out of tricks
                            throw new SyntaxException("In "+ file +": Parameter " + (index +1) +
                                    " on line " + line + " is not defined before it is used");
                        else{
                            // new desireable behaviour: Accept integers on places of REALs,
                            // consider REAL a more extensive type
                            /*
                            if (paramVar.isInteger())
                                throw new SyntaxException("In "+ file +": Parameter " + (index +1) +
                                        " on line " + line + " is not of type REAL");
                             */
                            dist.setParameter(index, paramVar);
                            paramVar.addAffected(toAdd);
                            toAdd.addDependence(paramVar);
                        }
                    }
                    
                }
                if (toParse == null){ // if we are going to terminate after this...
                    if (index+1 != dist.getNumberOfParameters())
                        throw new SyntaxException("In "+ file +": Too few parameters on line "+ line);
                }
                // if we didn't exit yet, the param is good, let's add the
                // String too
                dist.setParameterString(index, param);
                
                index++;
            }
            
        } else if(toParse.charAt(0) == '=') {
            toParse = toParse.substring(1).trim();
            toAdd.setFunctional(true);
            
            String equation = toParse;
            equation = equation.replaceAll("\\p{Space}+", "");
            
            ArrayList<String> tempString;
            ArrayList<Variable> variables;
            
            
            // now we have
            // EXP(alpha * 12 / LOG(beta - 1.0 * gamma)) ...
            // we are interested in parts alpha, beta, gamma and their starting indices.
            // IF I understand correctly, variable names are
            // the only part here written in lowercase a-z.
            p = Pattern.compile("[a-z][a-z0-9]*");
            m = p.matcher(equation);
            int last = 0;
            int index = 0;
            
            Pattern sum = Pattern.compile("SUM\\(\\&[a-z][a-z0-9]*\\)");
            Pattern count = Pattern.compile("COUNT\\(\\&[a-z][a-z0-9]*\\)");
            
            Matcher sumM = sum.matcher(equation);
            Matcher countM = count.matcher(equation);
            
            tempString = new ArrayList<String>();
            variables = new ArrayList<Variable>();
            boolean add = true;
            
            if (!sumM.matches() && equation.contains("SUM("))
                throw new SyntaxException("In " + file + ": 'SUM(&varname)' must be the only " +
                        "element of an equation if it is used (line " + line +")");
            if (!countM.matches() && equation.contains("COUNT("))
                throw new SyntaxException("In " + file + ": 'COUNT(&varname)' must be the only " +
                        "element of an equation if it is used (line " + line +")");
            
            // toAdd is spatial iff the equation matches
            // either one of the spatial statements.
            toAdd.setSpatial(sumM.matches() || countM.matches());
            
            while (m.find()){
                add = true;
                tempString.add(equation.substring(last, m.start()));
                last = m.end();
                String newvar = m.group();
                index++;
                tempString.add(newvar);
                for (Variable var : variables){
                    if (var.getName().equals(newvar))
                        add = false;
                }
                if (add){
                    if ((equation.indexOf('&') != -1) && equation.indexOf('&') == m.start() -1 ){
                        // if the variable is preceded with a '&'
                        // then it is to be checked later.
                        toBeChecked.add(toAdd);
                        spatAffectedMap.put(toAdd, newvar);
                    } else{
                        Variable dep = variableMapper.get(newvar);
                        if (dep == null)
                            throw new SyntaxException("In "+ file +": Variable " + index + " on line " + line
                                    + " is not defined before it is used");
                        else{
                            toAdd.addDependence(dep);
                            dep.addAffected(toAdd);
                            variables.add(dep);
                        }
                    }
                }
            }
            // note to self: last is the offset AFTER the last match.
            if (last != equation.length())
                tempString.add(equation.substring(last));
            
            toAdd.setEquation(new Equation(tempString.toArray(new String[tempString.size()]),
                    variables.toArray(new Variable[variables.size()])));
            
        }
        
        return toAdd;
    }
    
    /** Reads an entity section in the model file and saves the information into the structures provided.
     *
     *
     */
    
    private static void readEntity(File file, LinkedList<Entity> entityList,
            HashMap<String, Entity> entityMapper, HashMap<String, Variable> variableMapper,
            String header, Scanner reader, DistributionFactory fact)
            throws SyntaxException, MissingDistributionException{
        Entity toAdd = new Entity();
        
        // the line we have is an ENTITY definition line.
        //let's parse it first, and then
        
        /* header is:
           square
           or
           square, "squares.txt"
           or
           square, "ehkjotain.txt", "jotain.txt"
           or
           squarebird, "observations.txt", combines(square, bird)
           it can have a { at the line end, or not. let's check that first.
         */
        boolean curly = header.indexOf('{') == header.length() -1;
        int comma = header.indexOf(',');
        
        // name , "sth" [, "sth1" || combines(sth2, sth3)]
        
        String name = null;
        
        if (comma > 0)
            name = header.substring(0, header.indexOf(',')).trim();
        else {
            throw new SyntaxException("In " + file + ": name, \"datafile\" expected on line "+ line);
        }
        
        
        toAdd.setName(name);
        
        String dataFile = null;
        String spatialFile = null;
        String x = null;
        String y = null;
        
        comma = header.indexOf(',');
        
        header = header.substring(comma +1).trim();
        // "sth" , "sth1" || combines(sth2,sth3)
        int caption = header.indexOf('"');
        if (caption < 0)
            throw new SyntaxException("In " + file + ": \"datafile\" expected on line " + line);
        header = header.substring(caption +1).trim();
        dataFile = header.substring(0, header.indexOf('"')).trim();
        header = header.substring(header.indexOf('"') +1).trim();
        if (dataFile.length() == 0)
            dataFile = null;
        toAdd.setDataFile(dataFile);
        
        
        
        //  now dataFile is either null or "something.txt"
        
        /*
          header is:
          ""
          or
          , "sth1"
          or
          , combines(sth2, sth3)
         */
        
        comma = header.indexOf(',');
        if (comma >= 0 && header.length() > 0){ // something more, so header is
            header = header.substring(header.indexOf(',') +1).trim();
            /*
              "sth1"
              or
              combines(sth2, sth3)
             */
            if (header.startsWith("combines")){
                toAdd.setMatrix(true);
                int lparen = header.indexOf('(');
                int rparen = header.indexOf(')');
                if (lparen < 0 || rparen < 0)
                    throw new SyntaxException("Incorrect parentheses after combines-keyword on line "
                            + line);
                
                header = header.substring(lparen +1, rparen).trim();
                comma = header.indexOf(',');
                if (comma < 0)
                    throw new SyntaxException("combines(entity1, entity2) expected on line " + line);
                
                y = header.substring(0, comma).trim();
                x = header.substring(comma +1).trim();
                Entity yEntity = entityMapper.get(y);
                Entity xEntity = entityMapper.get(x);
                
                if (yEntity == null)
                    throw new SyntaxException("First entity for combines on line " + line +
                            " is not defined before it is used");
                if (xEntity == null)
                    throw new SyntaxException("Second entity for combines on line " + line +
                            " is not defined before it is used");
                
                toAdd.setYCoordinate(yEntity);
                toAdd.setXCoordinate(xEntity);
                
                
                
            } else{ // not a matrix
                toAdd.setMatrix(false);
                caption = header.indexOf('"');
                if (caption < 0)
                    throw new SyntaxException("\"spatialfile\" expected on line " + line);
                else{
                    header = header.substring(caption +1);
                    caption = header.indexOf('"');
                    if (caption < 0)
                        throw new SyntaxException("ending \" missing for \"spatialfile\" on line " + line);
                    spatialFile = header.substring(0, caption);
                    toAdd.setSpatialMatrixFile(spatialFile);
                }
            }
        }
        
        // now Entity properties name, dataFile, spatialFile, isMatrix, x and y
        // have been set, if applicable.
        
        // continue with the entity structure.
        String temp;
        while (reader.hasNextLine()){
            temp = reader.nextLine();
            
            line++;
            temp = temp.trim();
            
            if (temp.length() == 0 || temp.charAt(0) == '#')
                continue;
            
            if (temp.charAt(temp.length() -1) == '}')
                break;
            
            // now we have a non-comment non-zero line.
            // let's do something brilliant with it.
            
            if(!curly)
                if (temp.charAt(0) == '{'){
                curly = true;
                temp = temp.substring(1);
                if (temp.length() == 0 || temp.charAt(0) == '#')
                    continue;
                }
            
            if (curly){
                // do the work
                
                // add the Variables to the Entity
                if (temp.startsWith("REAL")){
                    temp = temp.substring(4).trim();
                    Variable var = parseVariable(file, false, temp, variableMapper, fact);
                    toAdd.addVariable(var);
                    variableMapper.put(var.getName(), var);
                    
                } else if (temp.startsWith("INTEGER")){
                    temp = temp.substring(7).trim();
                    Variable var = parseVariable(file, true, temp, variableMapper, fact);
                    toAdd.addVariable(var);
                    variableMapper.put(var.getName(), var);
                } else
                    throw new SyntaxException("Not a variable inside an entity on line " + line);
            }
            
        }
        
        // link the Entity to the Variables
        for (Variable toLinkUp : toAdd.getVariableList())
            toLinkUp.setEntity(toAdd);
        
        // put the Entity in the map
        //System.out.println("Adding: " + toAdd.getName());
        entityMapper.put(toAdd.getName(), toAdd);
        // link the Entity to the Model
        entityList.add(toAdd);
    }
    
    
    /** Reads a simulation parameter file that gives burn-in and thinning. Returns {burn-in, thinning}.
     *
     *
     */
    
    private static int[] readSimulation(File file)  throws SyntaxException, IOException{
        // reads stuff from the file and return burnIn and thinning. (int[2])
        /*
          ?? simulation
          ? burn-in
          100
          ? thinning
          100
         */
        return readUnifiedFormatFile(file, "?? simulation", "? burn-in", "? thinning");
    }
    
    /** Reads the update strategy file, saves individual updates (if any) into the variables in the maps,
     @return {String updateStrategy, int iterations}
     *
     *
     */
    private static Object[] readUpdate(File file, HashMap<String, Variable> variableMapper,
            String updateStrategy)  throws SyntaxException, IOException{
        /* reads the file, parses updateStrategy into the String variable and
           returns the number of iterations if updateStrategy == sequential.
           else returns -1, and reads the variables' update numbers.
         */
        
        Object[] toReturn = new Object[2];
        int iterations = -1;
        Scanner reader = new Scanner(file);
        
        line = 0;
        String temp = null;
        updateStrategy = null;
        boolean rightFile = false;
        boolean one = false;
        boolean strategy = false;
        String fileHeader = "?? update";
        String headerOne = "? strategy";
        
        while (reader.hasNextLine()){
            temp = reader.nextLine();
            
            line++;
            temp = temp.trim();
            
            if (temp.length() == 0 || temp.charAt(0) == '#')
                continue;
            
            if (temp.equals(fileHeader)){
                rightFile = true;
                continue;
            }
            
            if (rightFile){
                if (temp.equals(headerOne)){
                    one = true;
                    continue;
                }
                
                if (one && updateStrategy == null){ // updatestrategy header found, read that
                    updateStrategy = temp;
                    if (updateStrategy.equals("sequential"))
                        iterations = readUpdateValues(true, file, reader, variableMapper);
                    else if (updateStrategy.equals("random"))
                        iterations = readUpdateValues(false, file, reader, variableMapper);
                    else
                        throw new SyntaxException("In "+ file + ": Unknown update strategy on line "
                                + line);
                    
                    continue;
                } else {
                    throw new SyntaxException("In "+ file + ": Unknown identifier on line " + line);
                }
            } else throw new SyntaxException(("In "+ file + ": File identifier does not match " + fileHeader));
        }
        reader.close();
        
        toReturn[0] = updateStrategy;
        toReturn[1] = iterations;
        return toReturn;
    }
    
    /** Helper for readUpdate.
     *
     */
    private static int readUpdateValues(boolean sequential, File file, Scanner reader,
            HashMap <String, Variable> variableMapper)   throws SyntaxException {
        boolean seqHeader = false;
        String header = null;
        String temp = null;
        int iterations = -1;
        ArrayList<Variable> varArray = new ArrayList<Variable>(variableMapper.values());
        
        while (reader.hasNextLine()){
            temp = reader.nextLine();
            
            line++;
            temp = temp.trim();
            
            if (temp.length() == 0 || temp.charAt(0) == '#')
                continue;
            if (sequential){
                if(temp.equals("? iterations")){
                    seqHeader = true;
                    continue;
                }
                
                if(seqHeader && iterations == -1){ // hasn't been set yet
                    try{
                        iterations = Integer.parseInt(temp);
                    }catch(NumberFormatException e){
                        throw new SyntaxException("In "+ file + ": Not an integer on line " + line);
                    }
                }
            } else{
                // the hard part:
                // expecting:
                // ? updates
                // 500
                // ? varname
                // 700
                // ? varname
                // 600
                // ...
                if (temp.startsWith("?")){
                    header = temp;
                    continue;
                }
                if (header != null){
                    int updates = 0;
                    try {
                        updates = Integer.parseInt(temp);
                    }catch(NumberFormatException e){
                        throw new SyntaxException("In "+ file + ": Not an integer on line " + line);
                    }
                    if (header.equals("? updates")){
                        for (Variable var : varArray){
                            var.setUpdates(updates);
                        }
                    } else{
                        String name = header.substring(1).trim();
                        if (name.indexOf(' ') > 0)
                            name = name.substring(0, name.indexOf(' ')).trim();
                        // if it has "all" or something else after the name
                        Variable var = variableMapper.get(name);
                        if (var != null)
                            var.setUpdates(updates);
                        else
                            throw new SyntaxException("In "+ file + ": Unknown variable name on line "
                                    + line);
                    }
                    
                }
            }
        }
        
        return iterations;
    }
    
    /** Helper for readSimulation.
     *
     */
    private static int[] readUnifiedFormatFile(File file, String fileHeader, String headerOne,
            String headerTwo) throws SyntaxException, IOException {
        Scanner reader = new Scanner(file);
        
        line = 0;
        String temp = null;
        boolean rightFile = false;
        boolean one = false;
        boolean two = false;
        int[] toReturn = new int[2];
        
        while (reader.hasNextLine()){
            temp = reader.nextLine();
            
            line++;
            temp = temp.trim();
            
            if (temp.length() == 0 || temp.charAt(0) == '#')
                continue;
            
            if (temp.equals(fileHeader)){
                rightFile = true;
                continue;
            }
            
            if (rightFile){
                if (temp.equals(headerOne)){
                    one = true;
                    two = false;
                    continue;
                }
                
                if (temp.equals(headerTwo)){
                    one = false;
                    two = true;
                    continue;
                }
                
                if (one){
                    try{
                        toReturn[0] = Integer.parseInt(temp);
                    }catch(NumberFormatException e){
                        throw new SyntaxException("In "+ file + ": Not an integer on line " + line);
                    }
                }
                
                if (two){
                    try{
                        toReturn[1] = Integer.parseInt(temp);
                    }catch(NumberFormatException e){
                        throw new SyntaxException("In "+ file + ": Not an integer on line " + line);
                    }
                }
                
            } else throw new SyntaxException(("In "+ file + ": File identifier does not match " + fileHeader));
        }
        reader.close();
        
        return toReturn;
    }
    
    /** Reads the parameter names to output -file and sets isPrinted-flags.
     *
     */
    private static String[] readOutput(File file, HashMap<String, Variable> variableMapper)
    throws SyntaxException, IOException{
        /*
        reads the variables to be output from the file,
        sets variable.setPrinted() for those variables.
         */
        
        /* expecting:
           ?? output
           ? alpha
           ? x
         */
        
        String[] toReturn = new String[2];
        Scanner reader = new Scanner(file);
        int line = 0;
        String header = null;
        String temp = null;
        boolean rightFile = false;
        String fileHeader = "?? output";
        
        while (reader.hasNextLine()){
            temp = reader.nextLine();
            
            line++;
            temp = temp.trim();
            
            if (temp.length() == 0 || temp.charAt(0) == '#')
                continue;
            
            if (!rightFile && temp.startsWith(fileHeader)){
                rightFile = true;
                temp = temp.substring(9); // after ?? output
                if (temp.length() > 0){
                    StringTokenizer tkn = new StringTokenizer(temp, " ");
                    if (tkn.countTokens() != 2)
                        throw new SyntaxException("In "+ file + ": ?? output outputFileName " +
                                "summaryFileName expected on line "+ line);
                    toReturn[0] = tkn.nextToken();
                    toReturn[1] = tkn.nextToken();
                    
                } else
                    throw new SyntaxException("In "+ file + ": '?? output outputFileName summaryFileName' " +
                            "expected on line "+ line);
                continue;
            }
            if (rightFile){
                if (temp.startsWith("?")){
                    Variable var = null;
                    String all = null;
                    String name = temp.substring(1).trim();
                    // if it has all after it..
                    if (name.indexOf(' ') > 0){
                        all = name.substring(name.indexOf(' ')).trim();
                        name = name.substring(0, name.indexOf(' ')).trim();
                        if (!all.equals(all))
                            throw new SyntaxException("In "+ file + ": Expecting '? varname all' or " +
                                    "'? varname' on line "+ line);
                    }
                    var = variableMapper.get(name);
                    if (var != null){
                        // if the user has not specified 'all' for a nonglobal
                        if (all == null && var.getEntity() != null){
                            throw new SyntaxException("In "+ file + ": Expecting '? varname all' " +
                                    "for a non-global variable on line "+ line);
                        }
                        var.setPrinted();
                    } else
                        throw new SyntaxException("In "+ file + ": Unknown variable name on line " + line);
                }
            } else
                throw new SyntaxException(("In "+ file + ": File identifier does not match '" +
                        fileHeader +"'"));
        }
        reader.close();
        return toReturn;
    }
    
    /** Reads the proposal distributions/strategies file, sets them for specified variables.
     *
     */
    private static void readProposal(File file, HashMap<String, Variable> variableMapper,
            DistributionFactory fact) throws SyntaxException, IOException,
            MissingDistributionException{
        /* expecting:
           ?? proposal distributions
           ? alpha
           continuous_uniform(-1.0, 1.0)
           ? x all
           discrete_uniform(0,1)
         */
        Scanner reader = new Scanner(file);
        int line = 0;
        String header = null;
        String temp = null;
        boolean rightFile = false;
        String fileHeader = "?? proposal distributions";
        
        while (reader.hasNextLine()){
            temp = reader.nextLine();
            
            line++;
            temp = temp.trim();
            
            if (temp.length() == 0 || temp.charAt(0) == '#')
                continue;
            
            if (!rightFile && temp.equals(fileHeader)){
                rightFile = true;
                continue;
            }
            if (rightFile){
                if (temp.startsWith("?")){
                    header = temp;
                    continue;
                }
                if (header != null){
                    String name = header.substring(1).trim();
                    // if it has "all" or other stuff after it
                    if (name.indexOf(' ') > 0)
                        name = name.substring(0, name.indexOf(' ')).trim();
                    Variable var = variableMapper.get(name);
                    if (var != null){
                        // set the proposal distribution on this line,
                        // should match a distribution name.
                        // This is copied & modified from parserVariable.
                        // allowing proposal distribution parameters to
                        // be variables only means uncommenting a few lines.
                        // extendability.
                        
                        String distStr = temp.substring(0, temp.indexOf('(')).trim();
                        Distribution dist = fact.getDistribution(distStr);
                        // this might throw a MissingDistributionException.
                        
                        var.setProposal(dist);
                        String strategy = temp.substring(temp.lastIndexOf(')') + 1).trim();
                        if (strategy.equalsIgnoreCase("RW"))
                            var.setStrategy("random walk");
                        else
                            var.setStrategy("fixed");
                        temp = temp.substring(temp.indexOf('(') +1, temp.lastIndexOf(')')).trim();
                        // Now we have "param1, param2, ..., paramn"
                        
                        int index = 0;
                        while (temp != null){
                            int comma = temp.indexOf(',');
                            String param = null;
                            if (comma > 0) {
                                param = temp.substring(0, comma).trim();
                                // there is a comma, so the next parameter should be before it
                                // for the next round:
                                temp = temp.substring(comma +1).trim();
                            } else {
                                param = temp.trim();
                                // no comma found, so this is the last parameter
                                
                                // to terminate the loop
                                temp = null;
                                
                                // are there too few parameters in the file?
                                if (index+1 != dist.getNumberOfParameters())
                                    throw new SyntaxException("Too few parameters on line "+ line);
                            }
                            
                            // are there too many parameters in the file?
                            if (dist.getNumberOfParameters() <= index)
                                throw new SyntaxException("Too many parameters on line "+ line);
                            boolean integer = dist.isInteger(index);
                            if (integer){
                                try{
                                    int paramInt = Integer.parseInt(param);
                                    dist.setParameter(index, paramInt);
                                    // it's a good integer, no need to check for anything else
                                    
                                } catch(NumberFormatException e){
                                    // okay, it's not an Integer.
                                    // it may still be a Variable reference, let's not cry yet
                                    /*
                                        Variable paramVar = variableMapper.get(param);
                                    if (paramVar == null) // run out of tricks
                                     throw new SyntaxException("Parameter " + (index +1) + " on line " +
                                     line + " Is not defined before it is used");
                                     else{
                                        dist.setParameter(index, paramVar);
                                         paramVar.addAffected(var);
                                         var.addDependence(paramVar);
                                     */
                                    // and comment this:
                                    throw new SyntaxException("Proposal distribution has a non-constant " +
                                            "parameter " + (index +1) +" on line "+ line);
                                }
                            } else{
                                // not an integer, so it's REAL
                                try {
                                    double paramDouble = Double.parseDouble(param);
                                    dist.setParameter(index, paramDouble);
                                    // it's a good double, continue
                                    
                                } catch (NumberFormatException nfe){
                                    // okay, not a good double.
                                    
                                    // is it a Variable?
                                    /*
                                      Variable paramVar = variableMapper.get(param);
                                      if (paramVar == null) // run out of tricks
                                      throw new SyntaxException("Parameter " + (index +1) + " on line "
                                     + line + " Is not defined before it is used");
                                      else{
                                      dist.setParameter(index, paramVar);
                                        // paramVar.addAffected(var);
                                        // var.addDependence(paramVar);
                                     
                                     */
                                    
                                    // and comment this
                                    throw new SyntaxException("Proposal distribution has a non-constant " +
                                            "parameter " + (index +1) +" on line "+ line);
                                }
                            }
                            
                            // if we didn't exit yet, the param is good, let's add the
                            // String too
                            dist.setParameterString(index, param);
                            
                            index++;
                        }
                        
                    } else
                        throw new SyntaxException("In "+ file + ": Unknown variable name on line " + line);
                }
            } else
                throw new SyntaxException("In "+ file + ": File identifier does not match " + fileHeader);
        }
        reader.close();
        
    }
    
    /** Reads through entities' data files to see if they are correct. builds the missing values matrix files.
     *
     */
    private static void readEntityData(LinkedList<Entity> entityList) throws IOException,
            SyntaxException{
        /*
          Iterate through Entities, all their variables that are data, and
          set the data length(s). Since one Entity only has one dataFile,
          there are either 0 <= Variables with column > 0,
          entailing that only the number of lines in the file counts
          (and that all the lines are of equal length)
          or one Variable
          with column = -1 (taking up the whole dataFile),
          entailing that the Entity's XCoordinate and YCoordinate
          should get the number of columns and number of lines
          as their sizes, respectively.
         */
        boolean sizeSet = false;
        for(Entity e: entityList){
            if(e.getDataFile() != null){
                readData(e);
            }
        }
    }
    
    /** Helper for readEntityData.
     *
     */
    private static void readData(Entity e) throws IOException, SyntaxException{
        /*
           reads the data for a single Entity and sets the fields:
           datavar.missingValues, datavar.belongsTo.size
           reads out the missing values and writes the datafile_missing.txt
           to be used by later methods (readInitialValues)
         */
        
        File dataFile = new File(e.getDataFile());
        String dataName = e.getDataFile();
        FileWriter data = null;
        FileWriter missing = null;
        
        if (dataName.indexOf('.') < 0)
            data = new FileWriter(new File(dataName + "_1.txt"), false);
        else
            data = new FileWriter(dataName.substring(0, dataName.indexOf('.')) + "_1.txt", false);
        
        if (dataName.indexOf('.') < 0)
            missing = new FileWriter(dataName + "_missing.txt", false);
        else
            missing = new FileWriter(dataName.substring(0, dataName.indexOf('.')) + "_missing.txt", false);
        Scanner reader = new Scanner(dataFile);
        int dataVars = 0;
        String temp = null;
        
        for(Variable var : e.getVariableList())
            if (var.isData())
                dataVars++;
        
        int[][] cols = new int[dataVars][2]; // {column, type} (0-int, 1-double)
        Variable[] dataArray = new Variable[dataVars];
        
        Variable[] temporary = e.getVariableList().toArray(new Variable[e.getVariableList().size()]);
        
        dataVars = 0;
        for(int i = 0; i < temporary.length; ++i) {
            if (temporary[i].isData()){
                dataArray[dataVars] = temporary[i];
                cols[dataVars][0] = temporary[i].getColumn();
                if (temporary[i].isInteger())
                    cols[dataVars][1] = 0;
                else
                    cols[dataVars][1] = 1;
                dataVars++;
            }
        }
        
        // is this dangerous? I guess not. have to take this up.
        for (int i = 0; i < cols.length -1; ++i){
            for (int j = i +1; j < cols.length; ++j){
                if (cols[i][0] == cols[j][0])
                    throw new SyntaxException("Entity "+ e.getName() + " has two variables with the same column number");
            }
        }
        
        
        line = 0;
        int dataline = 0;
        int lineLength = 0;
        
        while (reader.hasNextLine()){
            temp = reader.nextLine();
            
            line++;
            temp = temp.trim();
            
            if (temp.length() == 0 || temp.charAt(0) == '#')
                continue;
            
            StringTokenizer tkn = new StringTokenizer(temp, " ");
            if (dataline != 0 && lineLength != tkn.countTokens())
                  throw new SyntaxException("In " + dataFile + ": Line " + line + 
                          "is of different length than the others");
            lineLength = tkn.countTokens();
            e.setLineLength(lineLength);
            int column = 0;
            int index = 0;
            String token = null;
            while (tkn.hasMoreTokens()){
                index = 0;
                column++; // columns are numbered starting from 1
                token = tkn.nextToken();
                
                
                // there is exactly one variable that can span the whole matrix.
                while (index < cols.length-1 && cols[index][0] != -1 && cols[index][0] != column)
                    index++;
                
                if (cols[index][0] != -1 && cols[index][0] != column){
                    //  System.out.println("No variable has been defined to use column " + column);
                    // write a 0 on its place
                    data.write(0 + " ");
                    missing.write(1 + " ");
                } else{
                    // the variable that spans the whole matrix must be matched
                    // against every column on every line, logically.
                    if (cols[index][0] == column || cols[index][0] == -1){
                        //  System.out.println("PARSER DEBUG: matching variable on column " + column);
                        
                        if (cols[index][1] == 0){ // integer
                            //    System.out.println("PARSER DEBUG: (integer)");
                            if (token.equals("*")){
                                // write a 0 on its place
                                data.write(0 + " ");
                                missing.write(1 + " ");
                                dataArray[index].incrementMissingValues();
                            } else{
                                try{
                                    Integer.parseInt(token);
                                    data.write(token + " ");
                                    missing.write(0 + " ");
                                }catch(NumberFormatException nfe){
                                    throw new SyntaxException("In " + dataFile + ": Value on line "+ line +
                                            " column " + column + " should be an integer");
                                }
                            }
                        } else{ // decimal(real)
                            //  System.out.println("PARSER DEBUG: (real)");
                            if (token.equals("*")){
                                // write a 0.0 on its place
                                data.write(0.0 + " ");
                                missing.write(1 + " ");
                                dataArray[index].incrementMissingValues();
                            } else{
                                try{
                                    Double.parseDouble(token);
                                    data.write(token + " ");
                                    missing.write(0 + " ");
                                }catch(NumberFormatException nfe){
                                    throw new SyntaxException("In " + dataFile + ": Value on line "+ line +
                                            " column " + column + " should be a decimal number");
                                }
                            }
                        }
                    }
                }
                
            }
            missing.write("\n");
            data.write("\n");
            
            dataline++; // do not count comment lines.
            
        }
        reader.close();
        missing.close();
        data.close();
        
        for (Variable v: dataArray){
            if (v.getMissingValueCount() > 0){
                if (v.getDistribution() == null)
                    throw new SyntaxException("Variable "+ v.getName() + " has missing values but " +
                            "has no distribution");
                
                if (v.getProposal() == null)
                    throw new SyntaxException("Variable "+ v.getName() + " has missing values but " +
                            "has no proposal distribution");
            }
        }
        
        
        if (e.isMatrix()){
            
            e.getYCoordinate().setSize(dataline);
            e.getXCoordinate().setSize(lineLength);
            e.setSize(-1);
        } else{
            e.setSize(dataline);
        }
    }
    
    /** Reads the initial values definition file and checks that the format is correct and all missing values of data are present.
     *
     */
    private static void readInitialValues(File file, HashMap<String, Variable> variableMapper)
    throws IOException, SyntaxException{
        /*
          Reads through every Entity,
          checks every non-functional Variable that is not data
          that is has getEntity().getSize() or if Entity.isMatrix()
          getEntity().getYCoordinate().getSize() *
          getEntity.getXCoordinate().getSize() initial values in the file.
          Checks also for each non-functional Variable that is data and
          has missingValueCount > 0
          that it has matching indices of initial values
          that are set as missing
          in the getEntity().dataFile_missing.txt.
         */
        
        /*
          expecting:
          ?? initial values
          ? alpha
          0.5
          ? x 1:21 1:4
          1 2 4 5
          1 2 4 6
          .......
         */
        
        Scanner reader = new Scanner(file);
        String temp = null;
        line = 0;
        String fileHeader = "?? initial values";
        String header = null;
        boolean rightFile = false;
        
        
        HashMap<Variable, boolean[][]> missingMap = new HashMap<Variable, boolean[][]>();
        // every index of the boolean[][] is information if this is found. (false == missing)
        // every variable has got a map like this of its data.
        
        ArrayList <Variable> varList = new ArrayList<Variable>(variableMapper.values());
        
        for (Variable v : varList){
            if (v.isFunctional())
                continue;
            if (v.isData() && v.getMissingValueCount() == 0)
                continue;
            
            if (v.isData()){
                if (v.getEntity() == null){
                    // global variables cannot be data
                    throw new SyntaxException("Global variable "+ v.getName() + " has been defined " +
                            "as data");
                }
                
                // missing values matrix stuff
                Entity entity = v.getEntity();
                String dataFileName = entity.getDataFile();
                Scanner missread = new Scanner(new File(dataFileName));
                String misstemp = null;
                int dataline = 0;
                boolean[][] missing;
                if (entity.isMatrix())
                    missing = new boolean[entity.getYCoordinate().getSize()][entity.getXCoordinate().getSize()];
                else
                    missing = new boolean[entity.getSize()][1];
                
                while(missread.hasNextLine()){
                    misstemp = missread.nextLine().trim();
                    dataline++;
                    int column = 0;
                    String token = null;
                    StringTokenizer tkn = new StringTokenizer(misstemp, " ");
                    while (tkn.hasMoreTokens()){
                        column++;
                        token = tkn.nextToken();
                        if (token.equals("*"))
                            missing[dataline-1][column-1] = false; // missing
                        else
                            missing[dataline-1][column-1] = true; // found
                        
                    }
                }
                
                missingMap.put(v, missing);
            } else{
                if (v.getEntity() == null){
                    // global, one value
                    boolean[][] missing = new boolean[1][1];
                    missing[0][0] = false; // we are missing the one value that has no indices
                    missingMap.put(v, missing);
                    
                } else if (!v.getEntity().isMatrix()){
                    // one-dim
                    int limit = v.getEntity().getSize();
                    // missing every index to size, that is:
                    boolean[][] missing = new boolean[limit][1];
                    for (int i = 1; i <= limit; ++i)
                        missing[i-1][0] = false;
                    missingMap.put(v, missing);
                } else{
                    // two-dim
                    int ylimit = v.getEntity().getYCoordinate().getSize();
                    int xlimit = v.getEntity().getXCoordinate().getSize();
                    
                    // missing every index to size, that is:
                    boolean[][] missing = new boolean[ylimit][xlimit];
                    
                    for (int i = 1; i <= ylimit; ++i)
                        for (int j = 1; j <= xlimit; ++j)
                            missing[i-1][j-1] = false;
                    missingMap.put(v, missing);
                }
            }
        }
        
        while (reader.hasNextLine()){
            temp = reader.nextLine().trim();
            
            line++;
            
            if (temp.length() == 0 || temp.charAt(0) == '#')
                continue;
            
            if (temp.equals(fileHeader)){
                rightFile = true;
                continue;
            }
            
            if (rightFile){
                if (temp.startsWith("?")){
                    header = temp.substring(1).trim();
                    continue;
                }
                if (header != null){
                    // alpha [1:12 [1:54]]
                    String name = null;
                    if (header.indexOf(' ') != -1){
                        name = header.substring(0, header.indexOf(' ')).trim();
                        header = header.substring(header.indexOf(' ')+1).trim();
                    } else {
                        name = header.trim();
                        header = "";
                    }
                    // [1:12 [1:54]]
                    Variable var = variableMapper.get(name);
                    if (var == null)
                        throw new SyntaxException("In "+ file + ": Unknown variable name on line " + line);
                    else{
                        // cases: a global variable:
                        // a one-dimensional entity variable
                        // a two-dimensional entity variable
                        if (var.isFunctional())
                            throw new SyntaxException("In "+ file + ": Attempting to give an initial " +
                                    "value to a functional variable on line "+ line);
                        if (var.isData() && var.getMissingValueCount() == 0)
                            throw new SyntaxException("In "+ file + ": Attempting to give an initial " +
                                    "value to a data variable on line "+ line);
                        
                        if (var.getEntity() == null){
                            // global, expecting one value
                            if (header.length() > 0)
                                throw new SyntaxException("In "+ file + ": Unidentified characters " +
                                        "after variable name on line "+ line);
                            else{
                                if (var.isInteger()){
                                    try{
                                        Integer.parseInt(temp);
                                    }catch(NumberFormatException nfe){
                                        throw new SyntaxException("In " + file + ": Value on line "+
                                                line + " should be an integer");
                                    }
                                } else{
                                    try{
                                        Double.parseDouble(temp);
                                    }catch(NumberFormatException nfe){
                                        throw new SyntaxException("In " + file + ": Value on line "+
                                                line + " should be a decimal number");
                                    }
                                }
                                // all is well if we didn't exit yet, and we can:
                                missingMap.get(var)[0][0] = true;
                            }
                        } else if(!var.getEntity().isMatrix()){
                            // one-dim
                            if (header.indexOf(':') == -1)
                                throw new SyntaxException("In " + file + ": '? varname " +
                                        "startindex:endindex' " +
                                        "expected on line "+ line);
                            
                            StringTokenizer limt = new StringTokenizer(header, ": ");
                            String one = limt.nextToken();
                            String two = limt.nextToken();
                            int start = -1;
                            int end = -1;
                            try{
                                start = Integer.parseInt(one);
                                end = Integer.parseInt(two);
                                if (start <= 0 || end <= 0)
                                    throw new SyntaxException("In "+ file + ": Boundaries should start from 1 minimum " +
                                            "on line "+ line);
                            }catch(NumberFormatException nfe){
                                throw new SyntaxException("In " + file + ": Boundaries are not integers " +
                                        "on line "+ line);
                            }
                            
                            readMissing(file, var.isInteger(), start, end, missingMap.get(var),
                                    temp, reader);
                            // reads the values from start to end from temp and reader, checks stuff
                            // on missingmap to true ([index][2] = 1;)
                        } else {
                            // two-dim
                            StringTokenizer limt = new StringTokenizer(header, ": ");
                            if (header.indexOf(':') == -1 || header.indexOf(':') ==
                                    header.lastIndexOf(':'))
                                throw new SyntaxException("In " + file + ": '? varname " +
                                        "ystartindex:yendindex xstartindex:xendindex' expected on line "
                                        + line);
                            
                            String vone = limt.nextToken();
                            String vtwo = limt.nextToken();
                            String hone = limt.nextToken();
                            String htwo = limt.nextToken();
                            
                            int vstart = -1;
                            int vend = -1;
                            int hstart = -1;
                            int hend = -1;
                            try{
                                vstart = Integer.parseInt(vone);
                                vend = Integer.parseInt(vtwo);
                                hstart = Integer.parseInt(hone);
                                hend = Integer.parseInt(htwo);
                                if (vstart <= 0 || vend <= 0 || hstart <= 0 || hend <= 0)
                                    throw new SyntaxException("In "+ file + ": Boundaries should start from 1 minimum " +
                                            "on line "+ line);
                            }catch(NumberFormatException nfe){
                                throw new SyntaxException("In " + file + ": All boundaries are not " +
                                        "integers on line "+ line);
                            }
                            readMissing(file, var.isInteger(), vstart, vend, hstart, hend,
                                    missingMap.get(var), temp, reader);
                        }
                    }
                }
            } else
                throw new SyntaxException("In "+ file + ": File identifier does not match " + fileHeader);
            
        }
        
        for (Variable v: varList){
            if (v.isFunctional())
                continue;
            if (v.isData() && v.getMissingValueCount() == 0)
                continue;
            
            boolean[][] missing = missingMap.get(v);
            for (int i = 1; i < missing.length; ++i)
                for (int j = 1; j < missing[i].length; ++j)
                    if (missing[i-1][j-1] == false){
                if (v.getEntity() == null)
                    throw new SyntaxException("In "+ file + ": Variable "+ v.getName() + " is missing its initial value");
                else
                    throw new SyntaxException("In "+ file + ": Variable "+ v.getName() + " is missing an initial " +
                            "value for ("+ i + ", " + j+ ")");
                    }
        }
        
    }
    
    /** Helper for readInitialValues.
     *
     */
    private static void readMissing(File file, boolean integer, int start, int end,
            boolean[][] missing, String temp, Scanner reader)
            throws SyntaxException{
        readMissing(file, integer, start, end, -1, -1, missing, temp, reader);
    }
    
    /** Helper for readInitialValues.
     *
     */
    private static void readMissing(File file, boolean integer, int vstart, int vend,
            int hstart, int hend, boolean[][] missing, String temp, Scanner reader)
            throws SyntaxException{
        String looptemp = null;
        int dataline = vstart;
        
        while (dataline <= vend){
            if (temp != null){
                looptemp = temp;
                temp = null;
            } else{
                if (!reader.hasNextLine()){
                    if (hstart > 0 && hend > 0)
                        throw new SyntaxException("In "+ file +": Not enough initial values, expected "+ vstart +":"+ vend + " "+ hstart + ":" + hend);
                    else
                        throw new SyntaxException("In "+ file +": Not enough initial values, expected "+ vstart +":"+ vend);
                }
                looptemp = reader.nextLine().trim();
                line++;
            }
            StringTokenizer tkn = new StringTokenizer(looptemp, " ");
            
            int column = hstart;
            
            while (column <= hend){
                if (!tkn.hasMoreTokens()){
                    if (hstart > 0 && hend > 0)
                        throw new SyntaxException("In "+ file +": Not enough initial values, expected "+ vstart +":"+ vend + " "+ hstart + ":" + hend);
                    else
                        throw new SyntaxException("In "+ file +": Not enough initial values, expected "+ vstart +":"+ vend);
                }
                
                String token = tkn.nextToken();
                
                if (integer){
                    try{
                        Integer.parseInt(token);
                    }catch(NumberFormatException nfe){
                        throw new SyntaxException("In " + file + ": Value " + dataline +", " +
                                column + " on line "+ line + " should be an integer");
                    }
                } else{
                    try{
                        Double.parseDouble(token);
                    }catch(NumberFormatException nfe){
                        throw new SyntaxException("In " + file + ": Value " + dataline +", " +
                                column + " on line "+ line + " should be a decimal number");
                    }
                }
                
                // it is of the correct type now
                if (column == -1)
                    missing[dataline-1][0] = true;
                else
                    missing[dataline-1][column -1] = true;
                
                column++;
            }
           // System.out.println("data line " + dataline + " parsed");
            dataline++;
        }
        
    }
    
    /** Counts the maximum number of spatial neighbours in the model.
     *
     */
    private static int countNeighbours(LinkedList<Entity> entityList) throws SyntaxException,
            FileNotFoundException{
        int toReturn = 0;
        int spatialEntities = 0;
        String spatial = null;
        
        for (int i = 0; i < entityList.size(); ++i){
            Entity e = entityList.get(i);
            if (e.getSpatialMatrixFile() != null){
                spatial = e.getSpatialMatrixFile();
                spatialEntities++;
                
            }
        }
        if (spatialEntities > 1)
            throw new SyntaxException("More than one spatial entity defined in the model!");
        
        if (spatialEntities == 0)
            return toReturn;
        // returning zero, no spatial Entities
        
        // now we have exactly one spatial entity
        // and the matrix filename is at spatial.
        
        Scanner reader = new Scanner(new File(spatial));
        String temp = null;
        line = 0;
        int maxcount = 0;
        int lineLength = 0;
        int dataline = 0;
        
        while (reader.hasNextLine()){
            temp = reader.nextLine();
            
            line++;
            temp = temp.trim();
            
            if (temp.length() == 0 || temp.charAt(0) == '#')
                continue;
            
            StringTokenizer tkn = new StringTokenizer(temp, " ");
            if (dataline != 0 && lineLength != tkn.countTokens())
                throw new SyntaxException("In " + spatial + ": Line " + line + " is of different " +
                        "length than the others");
            lineLength = tkn.countTokens();
            String token = null;
            maxcount = 0;
            while (tkn.hasMoreTokens()){
                token = tkn.nextToken();
                
                if (token.equals("1"))
                    maxcount++;
            }
            if (maxcount > toReturn){
                
                toReturn = maxcount;
            }
            dataline++; // don't evaluate comment lines
            
        }
        
        return toReturn;
    }
    
    /** Runs every variable through a sanity check (see Variable.isOk())
     *
     
     */
    private static void checkVariablesAreOk(ArrayList<Variable> variableList)
    throws SyntaxException{
        for (int i = 0; i < variableList.size()-1; ++i){
            for (int j = i +1; j < variableList.size(); ++j){
                if (variableList.get(i).equals(variableList.get(j)))
                    throw new SyntaxException("Two variables with the same name defined in the model.");
            }
        }
        for(Variable v: variableList){
            if (v.getProposal() == null){
                if (!v.isFunctional() && !v.isData())
                    throw new SyntaxException("Stochastic variable "+v.getName() + " has no proposal " +
                            "distribution");
            }
            if(!v.isOk()){
                System.out.println(v);
                throw new SyntaxException("Variable "+v.getName() + " is NOT ok");
            }
        }
    }
    
}
